# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import numpy as np
import math
import datetime
import warnings

from .angle import Angle, _Angle
from .angleunit import radians, degrees, hours, arcsec
from . import util

class CelestialCoord(object):
    """This class defines a position on the celestial sphere, normally given by two angles,
    `ra` and `dec`.

    This class can be used to perform various calculations in spherical coordinates, such
    as finding the angular distance between two points in the sky, calculating the angles in
    spherical triangles, projecting from sky coordinates onto a Euclidean tangent plane, etc.

    **Initialization:**

        A `CelestialCoord` object is constructed from the right ascension and declination:

            :meth:`coord.CelestialCoord.__init__`

            >>> c = CelestialCoord(ra=12*hours, dec=31*degrees)
            >>> print(c)
            coord.CelestialCoord(3.14159265359 radians, 0.541052068118 radians)

    **Attributes:**

        A CelestialCoord has the following (read-only) attributes:

            :ra:        The right ascension (an Angle instance)
            :dec:       The declination (an Angle instance)

            >>> print(c.ra / degrees, c.dec / degrees)
            180.0 31.0

        In addition there is a convenience access property that returns ra and dec in radians.

            :rad:       A tuple (ra.rad, dec.rad)

            >>> print(c.rad)
            (3.141592653589793, 0.5410520681182421)

    **Spherical Geometry:**

        The basic spherical geometry operations are available to work with spherical triangles

        For three coordinates cA, cB, cC making a spherical triangle, one can calculate the
        sides and angles via:

            | :meth:`coord.CelestialCoord.distanceTo`
            | :meth:`coord.CelestialCoord.angleBetween`

            >>> cA = CelestialCoord(0 * degrees, 0 * degrees)
            >>> cB = CelestialCoord(0 * degrees, 10 * degrees)
            >>> cC = CelestialCoord(10 * degrees, 0 * degrees)
            >>> a = cB.distanceTo(cC)
            >>> b = cC.distanceTo(cA)
            >>> c = cA.distanceTo(cB)
            >>> print(a / degrees, b / degrees, c / degrees)
            14.1060442606 10.0 10.0
            >>> A = cA.angleBetween(cB, cC)
            >>> B = cB.angleBetween(cC, cA)
            >>> C = cC.angleBetween(cA, cB)
            >>> print(A / degrees, B / degrees, C / degrees)
            90.0 45.4385485867 45.4385485867

    **Projections:**

        Local tangent plane projections of an area of the sky can be performed using the project
        method:

            :meth:`coord.CelestialCoord.project`

            >>> center = CelestialCoord(ra=10*hours, dec=30*degrees)
            >>> sky_coord = CelestialCoord(ra=10.5*hours, dec=31*degrees)
            >>> print(sky_coord)
            coord.CelestialCoord(2.74889357189 radians, 0.541052068118 radians)
            >>> u, v = center.project(sky_coord)
            >>> print(u / degrees, v / degrees)
            -6.45237127534 1.21794987289

        and back:

            :meth:`coord.CelestialCoord.deproject`

            >>> sky_coord = center.deproject(u,v)
            >>> print(sky_coord)
            coord.CelestialCoord(2.74889357189 radians, 0.541052068118 radians)

        where u and v are Angles and center and sky_coord are CelestialCoords.
    """
    def __init__(self, ra, dec=None):
        """
        :param ra:       The right ascension.  Must be an Angle instance.
        :param dec:      The declination.  Must be an Angle instance.
        """
        if isinstance(ra, CelestialCoord) and dec is None:
            # Copy constructor
            self._ra = ra._ra
            self._dec = ra._dec
            self._x = None
        elif ra is None or dec is None:
            raise TypeError("ra and dec are both required")
        elif not isinstance(ra, Angle):
            raise TypeError("ra must be a coord.Angle")
        elif not isinstance(dec, Angle):
            raise TypeError("dec must be a coord.Angle")
        elif dec/degrees > 90. or dec/degrees < -90.:
            raise ValueError("dec must be between -90 deg and +90 deg.")
        else:
            # Normal case
            self._ra = ra
            self._dec = dec
            self._x = None  # Indicate that x,y,z are not set yet.

    @property
    def ra(self):
        """A read-only attribute, giving the Right Ascension as an Angle"""
        return self._ra

    @property
    def dec(self):
        """A read-only attribute, giving the Declination as an Angle"""
        return self._dec

    @property
    def rad(self):
        """A convenience property, giving a tuple (ra.rad, dec.rad)
        """
        return (self._ra.rad, self._dec.rad)

    def _set_aux(self):
        if self._x is None:
            self._sindec, self._cosdec = self._dec.sincos()
            self._sinra, self._cosra = self._ra.sincos()
            self._x = self._cosdec * self._cosra
            self._y = self._cosdec * self._sinra
            self._z = self._sindec

    def get_xyz(self):
        """Get the (x,y,z) coordinates on the unit sphere corresponding to this (RA, Dec).

        .. math::

            x &= \\cos(dec) \\cos(ra)  \\\\
            y &= \\cos(dec) \\sin(ra)  \\\\
            z &= \\sin(dec)

        :returns: a tuple (x,y,z)
        """
        self._set_aux()
        return self._x, self._y, self._z

    @staticmethod
    def from_xyz(x, y, z):
        """Construct a CelestialCoord from a given (x,y,z) position in three dimensions.

        The 3D (x,y,z) position does not need to fall on the unit sphere.  The RA, Dec will
        be inferred from the relations:

        .. math::

            x &= r \\cos(dec) \\cos(ra)  \\\\
            y &= r \\cos(dec) \\sin(ra)  \\\\
            z &= r \\sin(dec)

        where :math:`r` is arbitrary.

        :param x:       The x position in 3 dimensions.  Corresponds to r cos(dec) cos(ra)
        :param y:       The y position in 3 dimensions.  Corresponds to r cos(dec) sin(ra)
        :param z:       The z position in 3 dimensions.  Corresponds to r sin(dec)

        :returns: a CelestialCoord instance
        """
        norm = np.sqrt(x*x + y*y + z*z)
        if norm == 0.:
            raise ValueError("CelestialCoord for position (0,0,0) is undefined.")
        ret = CelestialCoord.__new__(CelestialCoord)
        ret._x = x / norm
        ret._y = y / norm
        ret._z = z / norm
        ret._sindec = ret._z
        ret._cosdec = np.sqrt(ret._x*ret._x + ret._y*ret._y)
        if ret._cosdec == 0.:
            ret._sinra = 0.
            ret._cosra = 1.
        else:
            ret._sinra = ret._y / ret._cosdec
            ret._cosra = ret._x / ret._cosdec
        ret._ra = (np.arctan2(ret._sinra, ret._cosra) * radians).wrap(_Angle(math.pi))
        ret._dec = np.arctan2(ret._sindec, ret._cosdec) * radians
        return ret

    @staticmethod
    def radec_to_xyz(ra, dec, r=1.):
        """Convert ra, dec (in radians) to 3D x,y,z coordinates on the unit sphere.

        The connection between (ra,dec) and (x,y,z) are given by the following formulae:
        .. math::

            x &= r \\cos(dec) \\cos(ra)  \\\\
            y &= r \\cos(dec) \\sin(ra)  \\\\
            z &= r \\sin(dec)

        For a single ra,dec pair, the following are essentially equivalent:

            >>> ra = 12*hours/radians       # May be any angle measured
            >>> dec = 31*degrees/radians    # in radians

            >>> CelestialCoord.radec_to_xyz(ra, dec)
            (-0.85716730070211233, 1.0497271911386187e-16, 0.51503807491005416)

            >>> CelestialCoord(ra * radians, dec * radians).get_xyz()
            (-0.85716730070211233, 1.0497271911386187e-16, 0.51503807491005416)

        However, the advantage of this function is that the input values may be numpy
        arrays, in which case, the return values will also be numpy arrays.

        :param ra:      The right ascension(s) in radians. May be a numpy array.
        :param dec:     The declination(s) in radians. May be a numpy array.
        :param r:       The distance(s) from Earth (default 1.). May be a numpy array.

        :returns: x, y, z as a tuple.
        """
        cosdec = np.cos(dec)
        x = cosdec * np.cos(ra) * r
        y = cosdec * np.sin(ra) * r
        z = np.sin(dec) * r
        return x,y,z

    @staticmethod
    def xyz_to_radec(x, y, z, return_r=False):
        """Convert 3D x,y,z coordinates to ra, dec (in radians).

        The connection between (ra,dec) and (x,y,z) are given by the following formulae:
        .. math::

            x &= r \\cos(dec) \\cos(ra)  \\\\
            y &= r \\cos(dec) \\sin(ra)  \\\\
            z &= r \\sin(dec)

        For a single (x,y,z) position, the following are essentially equivalent:

            >>> x = 0.839       # May be any any 3D location
            >>> y = 0.123       # Not necessarily on unit sphere
            >>> z = 0.530

            >>> CelestialCoord.xyz_to_radec(x, y, z)
            (0.14556615088111796, 0.558616191048523)

            >>> c = CelestialCoord.from_xyz(x, y, z)
            >>> c.ra.rad, c.dec.rad
            (0.145566150881118, 0.558616191048523)

        However, the advantage of this function is that the input values may be numpy
        arrays, in which case, the return values will also be numpy arrays.

        :param x:       The x position(s) in 3 dimensions. May be a numpy array.
        :param y:       The y position(s) in 3 dimensions. May be a numpy array.
        :param z:       The z position(s) in 3 dimensions. May be a numpy array.
        :param return_r: Whether to return r as well as ra, dec. (default: False)

        :returns: ra, dec as a tuple.  Or if return_r is True, (ra, dec, r).
        """
        xy2 = x**2 + y**2
        ra = np.arctan2(y, x)
        # Note: We don't need arctan2, since always quadrant 1 or 4.
        #       Using plain arctan is slightly faster.  About 10% for the whole function.
        #       However, if any points have x=y=0, then this will raise a numpy warning.
        #       It still gives the right answer, but we catch and ignore the warning here.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",category=RuntimeWarning)
            dec = np.arctan(z/np.sqrt(xy2))
        if return_r:
            return ra, dec, np.sqrt(xy2 + z**2)
        else:
            return ra, dec

    def normal(self):
        """Return the coordinate in the "normal" convention of having 0 <= ra < 24 hours.

        This convention is not enforced on construction, so this function exists to make it
        easy to convert if desired.

        Functions such as `from_galactic` and `from_xyz` will return normal coordinates.
        """
        return _CelestialCoord(self.ra.wrap(_Angle(math.pi)), self.dec)

    def distanceTo(self, coord2):
        """Returns the great circle distance between this coord and another one.
        The return value is an Angle object

        :param coord2:      The CelestialCoordinate to calculate the distance to.

        :returns: the great circle distance to `coord2`.
        """
        # The easiest way to do this in a way that is stable for small separations
        # is to calculate the (x,y,z) position on the unit sphere corresponding to each
        # coordinate position.
        #
        # x = cos(dec) cos(ra)
        # y = cos(dec) sin(ra)
        # z = sin(dec)

        self._set_aux()
        coord2._set_aux()

        # The direct distance between the two points is
        #
        # d^2 = (x1-x2)^2 + (y1-y2)^2 + (z1-z2)^2

        dsq = (self._x-coord2._x)**2 + (self._y-coord2._y)**2 + (self._z-coord2._z)**2

        if dsq < 3.99:
            # (The usual case.  This formula is perfectly stable here.)
            # This direct distance can then be converted to a great circle distance via
            #
            # sin(theta/2) = d/2
            theta = 2. * math.asin(0.5 * math.sqrt(dsq))

        else:
            # Points are nearly antipodes where the accuracy of this formula starts to break down.
            # But in this case, the cross product provides an accurate distance.
            crosssq = ((coord2._y * self._z - coord2._z * self._y)**2 +
                       (coord2._z * self._x - coord2._x * self._z)**2 +
                       (coord2._x * self._y - coord2._y * self._x)**2)
            theta = math.pi - math.asin(math.sqrt(crosssq))

        return _Angle(theta)


    def _triple(self, coord2, coord3):
        """Compute the scalar triple product of the three vectors:

                (A x C). B = sina sinb sinC

        where C = self, A = coord2, B = coord3.  This is used by both angleBetween and area.
        (Although note that the triple product is invariant to the ordering modulo a sign.)
        """
        # Note, the scalar triple product, (AxC).B, is the determinant of the 3x3 matrix
        #     [ xA yA zA ]
        #     [ xC yC zC ]
        #     [ xB yB zB ]
        # Furthermore, it is more stable to calculate it that way than computing the cross
        # product by hand and then dotting it to the other vector.
        return np.linalg.det([ [ coord2._x, coord2._y, coord2._z ],
                               [   self._x,   self._y,   self._z ],
                               [ coord3._x, coord3._y, coord3._z ] ])


    def _alt_triple(self, coord2, coord3):
        """Compute a different triple product of the three vectors:

                (A x C). (B x C) = sina sinb cosC

        where C = self, A = coord2, B = coord3.  This is used by both angleBetween and area.
        """
        # We can simplify (AxC).(BxC) as follows:
        #     (A x C) . (B x C)
        #     = (C x (BxC)) . A         Rotation of triple product with (BxC) one of the vectors
        #     = ((C.C)B - (C.B)C) . A   Vector triple product identity
        #     = A.B - (A.C) (B.C)       C.C = 1
        # Dot products for nearby coordinates are not very accurate.  Better to use the distances
        # between the points: A.B = 1 - d_AB^2/2
        #     = 1 - d_AB^2/2 - (1-d_AC^2/2) (1-d_BC^2/2)
        #     = d_AC^2 / 2 + d_BC^2 / 2 - d_AB^2 / 2 - d_AC^2 d_BC^2 / 4
        dsq_AC = (self._x-coord2._x)**2 + (self._y-coord2._y)**2 + (self._z-coord2._z)**2
        dsq_BC = (self._x-coord3._x)**2 + (self._y-coord3._y)**2 + (self._z-coord3._z)**2
        dsq_AB = (coord3._x-coord2._x)**2 + (coord3._y-coord2._y)**2 + (coord3._z-coord2._z)**2
        return 0.5 * (dsq_AC + dsq_BC - dsq_AB - 0.5 * dsq_AC * dsq_BC)


    def angleBetween(self, coord2, coord3):
        """Find the open angle at the location of the current coord between `coord2` and `coord3`.

        The current coordinate along with the two other coordinates form a spherical triangle
        on the sky.  This function calculates the angle between the two sides at the location of
        the current coordinate.

        Note that this returns a signed angle.  The angle is positive if the sweep direction from
        `coord2` to `coord3` is counter-clockwise (as observed from Earth).  It is negative if
        the direction is clockwise.

        :param coord2:       A second CelestialCoord
        :param coord3:       A third CelestialCoord

        :returns: the angle between the great circles joining the other two coordinates to the
                  current coordinate.
        """
        # Call A = coord2, B = coord3, C = self
        # Then we are looking for the angle ACB.
        # If we treat each coord as a (x,y,z) vector, then we can use the following spherical
        # trig identities:
        #
        # (A x C) . B = sina sinb sinC
        # (A x C) . (B x C) = sina sinb cosC
        #
        # Then we can just use atan2 to find C, and atan2 automatically gets the sign right.
        # And we only need 1 trig call, assuming that x,y,z are already set up, which is often
        # the case.

        self._set_aux()
        coord2._set_aux()
        coord3._set_aux()

        sinC = self._triple(coord2, coord3)
        cosC = self._alt_triple(coord2, coord3)

        C = math.atan2(sinC, cosC)
        return _Angle(C)

    def area(self, coord2, coord3):
        """Find the area of a spherical triangle in steradians.

        The current coordinate along with the two other coordinates form a spherical triangle
        on the sky.  This function calculates the area of that spherical triangle, which is
        measured in steradians (i.e. surface area of the triangle on the unit sphere).

        :param coord2:       A second CelestialCoord
        :param coord3:       A third CelestialCoord

        :returns: the area in steradians of the given spherical triangle.
        """
        # The area of a spherical triangle is defined by the "spherical excess", E.
        # There are several formulae for E:
        #    (cf. http://en.wikipedia.org/wiki/Spherical_trigonometry#Area_and_spherical_excess)
        #
        # E = A + B + C - pi
        # tan(E/4) = sqrt(tan(s/2) tan((s-a)/2) tan((s-b)/2) tan((s-c)/2)
        # tan(E/2) = tan(a/2) tan(b/2) sin(C) / (1 + tan(a/2) tan(b/2) cos(C))
        #
        # We use the last formula, which is stable both for small triangles and ones that are
        # nearly degenerate (which the middle formula may have trouble with).
        #
        # Furthermore, we can use some of the math for angleBetween and distanceTo to simplify
        # this further:
        #
        # In angleBetween, we have formulae for sina sinb sinC and sina sinb cosC.
        # In distanceTo, we have formulae for sin(a/2) and sin(b/2).
        #
        # Define: F = sina sinb sinC
        #         G = sina sinb cosC
        #         da = 2 sin(a/2)
        #         db = 2 sin(b/2)
        #
        # tan(E/2) = sin(a/2) sin(b/2) sin(C) / (cos(a/2) cos(b/2) + sin(a/2) sin(b/2) cos(C))
        #          = sin(a) sin(b) sin(C) / (4 cos(a/2)^2 cos(b/2)^2 + sin(a) sin(b) cos(C))
        #          = F / (4 (1-sin(a/2)^2) (1-sin(b/2)^2) + G)
        #          = F / (4-da^2) (4-db^2)/4 + G)

        self._set_aux()
        coord2._set_aux()
        coord3._set_aux()

        F = self._triple(coord2, coord3)
        G = self._alt_triple(coord2, coord3)
        dasq = (self._x-coord2._x)**2 + (self._y-coord2._y)**2 + (self._z-coord2._z)**2
        dbsq = (self._x-coord3._x)**2 + (self._y-coord3._y)**2 + (self._z-coord3._z)**2

        tanEo2 = F / ( 0.25 * (4.-dasq) * (4.-dbsq) + G)
        E = 2. * math.atan( abs(tanEo2) )
        return E

    _valid_projections = [None, 'gnomonic', 'stereographic', 'lambert', 'postel']

    def project(self, coord2, projection=None):
        """Use the currect coord as the center point of a tangent plane projection to project
        the `coord2` coordinate onto that plane.

        This function return a tuple (u,v) in the Euclidean coordinate system defined by
        a tangent plane projection around the current coordinate, with +v pointing north and
        +u pointing west. (i.e. to the right on the sky if +v is up.)

        There are currently four options for the projection, which you can specify with the
        optional `projection` keyword argument:

            :gnomonic: [default] uses a gnomonic projection (i.e. a projection from the center of
                    the sphere, which has the property that all great circles become straight
                    lines.  For more information, see
                    http://mathworld.wolfram.com/GnomonicProjection.html
                    This is the usual TAN projection used by most FITS images.
            :stereographic: uses a stereographic proejection, which preserves angles, but
                    not area.  For more information, see
                    http://mathworld.wolfram.com/StereographicProjection.html
            :lambert: uses a Lambert azimuthal projection, which preserves area, but not angles.
                    For more information, see
                    http://mathworld.wolfram.com/LambertAzimuthalEqual-AreaProjection.html
            :postel: uses a Postel equidistant proejection, which preserves distances from
                    the projection point, but not area or angles.  For more information, see
                    http://mathworld.wolfram.com/AzimuthalEquidistantProjection.html

        The distance or angle errors increase with distance from the projection point of course.

        :param coord2:      The coordinate to project onto the tangent plane.
        :param projection:  The name of the projection to be used. [default: gnomonic, see above
                            for other options]

        :returns: (u,v) as Angle instances
        """
        if projection not in CelestialCoord._valid_projections:
            raise ValueError('Unknown projection: %s'%projection)

        self._set_aux()
        coord2._set_aux()

        # The core calculation is done in a helper function:
        u, v = self._project(coord2._cosra, coord2._sinra, coord2._cosdec, coord2._sindec,
                             projection)

        return u * radians, v * radians

    def project_rad(self, ra, dec, projection=None):
        """This is basically identical to the project() function except that the input `ra`, `dec`
        are given in radians rather than packaged as a CelestialCoord object and the returned
        u,v are given in radians.

        The main advantage to this is that it will work if `ra` and `dec` are NumPy arrays, in which
        case the output `u`, `v` will also be NumPy arrays.

        :param ra:          The right ascension in radians to project onto the tangent plane.
        :param dec:         The declination in radians to project onto the tangent plane.
        :param projection:  The name of the projection to be used. [default: gnomonic, see `project`
                            docstring for other options]

        :returns: (u,v) in radians
        """
        if projection not in CelestialCoord._valid_projections:
            raise ValueError('Unknown projection: %s'%projection)

        self._set_aux()

        cosra = np.cos(ra)
        sinra = np.sin(ra)
        cosdec = np.cos(dec)
        sindec = np.sin(dec)

        return self._project(cosra, sinra, cosdec, sindec, projection)

    def _project(self, cosra, sinra, cosdec, sindec, projection):
        # The equations are given at the above mathworld websites.  They are the same except
        # for the definition of k:
        #
        # x = k cos(dec) sin(ra-ra0)
        # y = k ( cos(dec0) sin(dec) - sin(dec0) cos(dec) cos(ra-ra0) )
        #
        # Lambert:
        #   k = sqrt( 2  / ( 1 + cos(c) ) )
        # Stereographic:
        #   k = 2 / ( 1 + cos(c) )
        # Gnomonic:
        #   k = 1 / cos(c)
        # Postel:
        #   k = c / sin(c)
        # where cos(c) = sin(dec0) sin(dec) + cos(dec0) cos(dec) cos(ra-ra0)

        # cos(dra) = cos(ra-ra0) = cos(ra0) cos(ra) + sin(ra0) sin(ra)
        cosdra = self._cosra * cosra
        cosdra += self._sinra * sinra

        # sin(dra) = -sin(ra - ra0)
        # Note: - sign here is to make +x correspond to -ra,
        #       so x increases for decreasing ra.
        #       East is to the left on the sky!
        # sin(dra) = -cos(ra0) sin(ra) + sin(ra0) cos(ra)
        sindra = self._sinra * cosra
        sindra -= self._cosra * sinra

        # Calculate k according to which projection we are using
        cosc = cosdec * cosdra
        cosc *= self._cosdec
        cosc += self._sindec * sindec
        if projection is None or projection[0] == 'g':
            k = 1. / cosc
        elif projection[0] == 's':
            k = 2. / (1. + cosc)
        elif projection[0] == 'l':
            k = np.sqrt( 2. / (1.+cosc) )
        elif cosc == 1.:
            k = 1.
        else:
            c = np.arccos(cosc)
            k = c / np.sin(c)

        # u = k * cosdec * sindra
        # v = k * ( self._cosdec * sindec - self._sindec * cosdec * cosdra )
        u = cosdec * sindra
        v = cosdec * cosdra
        v *= -self._sindec
        v += self._cosdec * sindec
        u *= k
        v *= k

        return u, v

    def deproject(self, u, v, projection=None):
        """Do the reverse process from the project() function.

        i.e. This takes in a position (u,v) and returns the corresponding celestial
        coordinate, using the current coordinate as the center point of the tangent plane
        projection.

        :param u:           The u position on the tangent plane to deproject (must be an Angle
                            instance)
        :param v:           The v position on the tangent plane to deproject (must be an Angle
                            instance)
        :param projection:  The name of the projection to be used. [default: gnomonic, see `project`
                            docstring for other options]

        :returns: the corresponding CelestialCoord for that position.
        """
        if projection not in CelestialCoord._valid_projections:
            raise ValueError('Unknown projection: %s'%projection)

        # Again, do the core calculations in a helper function
        ra, dec = self._deproject(u / radians, v / radians, projection)

        return CelestialCoord(_Angle(ra), _Angle(dec))

    def deproject_rad(self, u, v, projection=None):
        """This is basically identical to the deproject() function except that the output `ra`,
        `dec` are returned as a tuple (ra, dec) in radians rather than packaged as a CelestialCoord
        object and `u` and `v` are in radians rather than Angle instances.

        The main advantage to this is that it will work if `u` and `v` are NumPy arrays, in which
        case the output `ra`, `dec` will also be NumPy arrays.

        :param u:           The u position in radians on the tangent plane to deproject
        :param v:           The v position in radians on the tangent plane to deproject
        :param projection:  The name of the projection to be used. [default: gnomonic, see `project`
                            docstring for other options]

        :returns: the corresponding RA, Dec in radians
        """
        if projection not in CelestialCoord._valid_projections:
            raise ValueError('Unknown projection: %s'%projection)

        return self._deproject(u, v, projection)

    def _deproject(self, u, v, projection):
        # The inverse equations are also given at the same web sites:
        #
        # sin(dec) = cos(c) sin(dec0) + v sin(c) cos(dec0) / r
        # tan(ra-ra0) = u sin(c) / (r cos(dec0) cos(c) - v sin(dec0) sin(c))
        #
        # where
        #
        # r = sqrt(u^2+v^2)
        # c = tan^(-1)(r)     for gnomonic
        # c = 2 tan^(-1)(r/2) for stereographic
        # c = 2 sin^(-1)(r/2) for lambert
        # c = r               for postel

        # Note that we can rewrite the formulae as:
        #
        # sin(dec) = cos(c) sin(dec0) + v (sin(c)/r) cos(dec0)
        # tan(ra-ra0) = u (sin(c)/r) / (cos(dec0) cos(c) - v sin(dec0) (sin(c)/r))
        #
        # which means we only need cos(c) and sin(c)/r.  For most of the projections,
        # this saves us from having to take sqrt(rsq).

        rsq = u*u
        rsq += v*v
        if projection is None or projection[0] == 'g':
            # c = arctan(r)
            # cos(c) = 1 / sqrt(1+r^2)
            # sin(c) = r / sqrt(1+r^2)
            cosc = sinc_over_r = 1./np.sqrt(1.+rsq)
        elif projection[0] == 's':
            # c = 2 * arctan(r/2)
            # Some trig manipulations reveal:
            # cos(c) = (4-r^2) / (4+r^2)
            # sin(c) = 4r / (4+r^2)
            cosc = (4.-rsq) / (4.+rsq)
            sinc_over_r = 4. / (4.+rsq)
        elif projection[0] == 'l':
            # c = 2 * arcsin(r/2)
            # Some trig manipulations reveal:
            # cos(c) = 1 - r^2/2
            # sin(c) = r sqrt(4-r^2) / 2
            cosc = 1. - rsq/2.
            sinc_over_r = np.sqrt(4.-rsq) / 2.
        else:
            r = np.sqrt(rsq)
            cosc = np.cos(r)
            sinc_over_r = np.sinc(r/np.pi)

        # Compute sindec, tandra
        # Note: more efficient to use numpy op= as much as possible to avoid temporary arrays.
        self._set_aux()
        # sindec = cosc * self._sindec + v * sinc_over_r * self._cosdec
        sindec = v * sinc_over_r
        sindec *= self._cosdec
        sindec += cosc * self._sindec
        # Remember the - sign so +dra is -u.  East is left.
        tandra_num = u * sinc_over_r
        tandra_num *= -1.
        # tandra_denom = cosc * self._cosdec - v * sinc_over_r * self._sindec
        tandra_denom = v * sinc_over_r
        tandra_denom *= -self._sindec
        tandra_denom += cosc * self._cosdec

        dec = np.arcsin(sindec)
        ra = self.ra.rad + np.arctan2(tandra_num, tandra_denom)

        return ra, dec

    def jac_deproject(self, u, v, projection=None):
        """Return the jacobian of the deprojection.

        i.e. if the input position is (u,v) then the return matrix is

        .. math::

            J \\equiv \\begin{bmatrix} J00 & J01 \\\\ J10 & J11 \\end{bmatrix}
              = \\begin{bmatrix}
                d\\textrm{ra}/du \\cos(\\textrm{dec}) & d\\textrm{ra}/dv \\cos(\\textrm{dec})  \\\\
                        d\\textrm{dec}/du             &      d\\textrm{dec}/dv
                 \\end{bmatrix}

        :param u:           The u position (as an Angle instance) on the tangent plane
        :param v:           The v position (as an Angle instance) on the tangent plane
        :param projection:  The name of the projection to be used. [default: gnomonic, see `project`
                            docstring for other options]

        :returns: the Jacobian as a 2x2 numpy array [[J00, J01], [J10, J11]]
        """
        if projection not in CelestialCoord._valid_projections:
            raise ValueError('Unknown projection: %s'%projection)
        return self._jac_deproject(u.rad, v.rad, projection)

    def jac_deproject_rad(self, u, v, projection=None):
        """Equivalent to `jac_deproject`, but the inputs are in radians and may be numpy
        arrays.

        :param u:           The u position (in radians) on the tangent plane
        :param v:           The v position (in radians) on the tangent plane
        :param projection:  The name of the projection to be used. [default: gnomonic, see `project`
                            docstring for other options]

        :returns: the Jacobian as a 2x2 numpy array [[J00, J01], [J10, J11]]
        """
        if projection not in CelestialCoord._valid_projections:
            raise ValueError('Unknown projection: %s'%projection)
        return self._jac_deproject(u, v, projection)

    def _jac_deproject(self, u, v, projection):
        # sin(dec) = cos(c) sin(dec0) + v sin(c)/r cos(dec0)
        # tan(ra-ra0) = u sin(c)/r / (cos(dec0) cos(c) - v sin(dec0) sin(c)/r)
        #
        # d(sin(dec)) = cos(dec) ddec = s0 dc + (v ds + s dv) c0
        # dtan(ra-ra0) = sec^2(ra-ra0) dra
        #              = ( (u ds + s du) A - u s (dc c0 - (v ds + s dv) s0 ) )/A^2
        # where s = sin(c) / r
        #       c = cos(c)
        #       s0 = sin(dec0)
        #       c0 = cos(dec0)
        #       A = c c0 - v s s0

        rsq = u*u + v*v
        rsq1 = (u+1.e-4)**2 + v**2
        rsq2 = u**2 + (v+1.e-4)**2
        if projection is None or projection[0] == 'g':
            c = s = 1./np.sqrt(1.+rsq)
            s3 = s*s*s
            dcdu = dsdu = -u*s3
            dcdv = dsdv = -v*s3
        elif projection[0] == 's':
            s = 4. / (4.+rsq)
            c = 2.*s-1.
            ssq = s*s
            dcdu = -u * ssq
            dcdv = -v * ssq
            dsdu = 0.5*dcdu
            dsdv = 0.5*dcdv
        elif projection[0] == 'l':
            c = 1. - rsq/2.
            s = np.sqrt(4.-rsq) / 2.
            dcdu = -u
            dcdv = -v
            dsdu = -u/(4.*s)
            dsdv = -v/(4.*s)
        else:
            r = np.sqrt(rsq)
            if r == 0.:
                c = s = 1
                dcdu = -u
                dcdv = -v
                dsdu = dsdv = 0
            else:
                c = np.cos(r)
                s = np.sin(r)/r
                dcdu = -s*u
                dcdv = -s*v
                dsdu = (c-s)*u/rsq
                dsdv = (c-s)*v/rsq

        self._set_aux()
        s0 = self._sindec
        c0 = self._cosdec
        sindec = c * s0 + v * s * c0
        cosdec = np.sqrt(1.-sindec*sindec)
        dddu = ( s0 * dcdu + v * dsdu * c0 ) / cosdec
        dddv = ( s0 * dcdv + (v * dsdv + s) * c0 ) / cosdec

        tandra_num = u * s
        tandra_denom = c * c0 - v * s * s0
        # Note: A^2 sec^2(dra) = denom^2 (1 + tan^2(dra) = denom^2 + num^2
        A2sec2dra = tandra_denom**2 + tandra_num**2
        drdu = ((u * dsdu + s) * tandra_denom - u * s * ( dcdu * c0 - v * dsdu * s0 ))/A2sec2dra
        drdv = (u * dsdv * tandra_denom - u * s * ( dcdv * c0 - (v * dsdv + s) * s0 ))/A2sec2dra

        drdu *= cosdec
        drdv *= cosdec
        return np.array([[drdu, drdv], [dddu, dddv]])

    def precess(self, from_epoch, to_epoch):
        """This function precesses equatorial ra and dec from one epoch to another.
           It is adapted from a set of fortran subroutines based on (a) pages 30-34 of
           the Explanatory Supplement to the AE, (b) Lieske, et al. (1977) A&A 58, 1-16,
           and (c) Lieske (1979) A&A 73, 282-284.

        :param from_epoch:  The epoch of the current coordinate
        :param to_epoch:    The epoch of the returned precessed coordinate

        :returns: a CelestialCoord object corresponding to the precessed position.
        """
        if from_epoch == to_epoch: return self

        # t0, t below correspond to Lieske's big T and little T
        t0 = (from_epoch-2000.)/100.
        t = (to_epoch-from_epoch)/100.
        t02 = t0*t0
        t2 = t*t
        t3 = t2*t

        # a,b,c below correspond to Lieske's zeta_A, z_A and theta_A
        a = ( (2306.2181 + 1.39656*t0 - 0.000139*t02) * t +
              (0.30188 - 0.000344*t0) * t2 + 0.017998 * t3 ) * arcsec
        b = ( (2306.2181 + 1.39656*t0 - 0.000139*t02) * t +
              (1.09468 + 0.000066*t0) * t2 + 0.018203 * t3 ) * arcsec
        c = ( (2004.3109 - 0.85330*t0 - 0.000217*t02) * t +
              (-0.42665 - 0.000217*t0) * t2 - 0.041833 * t3 ) * arcsec
        sina, cosa = a.sincos()
        sinb, cosb = b.sincos()
        sinc, cosc = c.sincos()

        # This is the precession rotation matrix:
        xx = cosa*cosc*cosb - sina*sinb
        yx = -sina*cosc*cosb - cosa*sinb
        zx = -sinc*cosb
        xy = cosa*cosc*sinb + sina*cosb
        yy = -sina*cosc*sinb + cosa*cosb
        zy = -sinc*sinb
        xz = cosa*sinc
        yz = -sina*sinc
        zz = cosc

        # Perform the rotation:
        self._set_aux()
        x2 = xx*self._x + yx*self._y + zx*self._z
        y2 = xy*self._x + yy*self._y + zy*self._z
        z2 = xz*self._x + yz*self._y + zz*self._z

        return CelestialCoord.from_xyz(x2, y2, z2).normal()

    def galactic(self, epoch=2000.):
        """Get the longitude and latitude in galactic coordinates corresponding to this position.

        :param epoch:       The epoch of the current coordinate. [default: 2000.]

        :returns: the longitude and latitude as a tuple (el, b), given as Angle instances.
        """
        # cf. Lang, Astrophysical Formulae, page 13
        # cos(b) cos(el-33) = cos(dec) cos(ra-282.25)
        # cos(b) sin(el-33) = sin(dec) sin(62.6) + cos(dec) sin(ra-282.25) cos(62.6)
        #            sin(b) = sin(dec) cos(62.6) - cos(dec) sin(ra-282.25) sin(62.6)
        #
        # Those formulae were for the 1950 epoch.  The corresponding numbers for J2000 are:
        # (cf. https://arxiv.org/pdf/1010.3773.pdf)
        el0 = 32.93191857 * degrees
        r0 = 282.859481208 * degrees
        d0 = 62.8717488056 * degrees
        sind0, cosd0 = d0.sincos()

        sind, cosd = self.dec.sincos()
        sinr, cosr = (self.ra-r0).sincos()

        cbcl = cosd*cosr
        cbsl = sind*sind0 + cosd*sinr*cosd0
        sb = sind*cosd0 - cosd*sinr*sind0

        b = _Angle(math.asin(sb))
        el = (_Angle(math.atan2(cbsl,cbcl)) + el0).wrap(_Angle(math.pi))

        return (el, b)


    @staticmethod
    def from_galactic(el, b, epoch=2000.):
        """Create a CelestialCoord from the given galactic coordinates

        :param el:          The longitude in galactic coordinates (an Angle instance)
        :param b:           The latitude in galactic coordinates (an Angle instance)
        :param epoch:       The epoch of the returned coordinate. [default: 2000.]

        :returns: the CelestialCoord corresponding to these galactic coordinates.
        """
        el0 = 32.93191857 * degrees
        r0 = 282.859481208 * degrees
        d0 = 62.8717488056 * degrees
        sind0, cosd0 = d0.sincos()

        sinb, cosb = b.sincos()
        sinl, cosl = (el-el0).sincos()
        x1 = cosb*cosl
        y1 = cosb*sinl
        z1 = sinb

        x2 = x1
        y2 = y1 * cosd0 - z1 * sind0
        z2 = y1 * sind0 + z1 * cosd0

        temp = CelestialCoord.from_xyz(x2, y2, z2)
        return CelestialCoord(temp.ra + r0, temp.dec).normal()


    def ecliptic(self, epoch=2000., date=None):
        """Get the longitude and latitude in ecliptic coordinates corresponding to this position.

        The `epoch` parameter is used to get an accurate value for the (time-varying) obliquity of
        the ecliptic.  The formulae for this are quite straightforward.  It requires just a single
        parameter for the transformation, the obliquity of the ecliptic (the Earth's axial tilt).

        :param epoch:       The epoch to be used for estimating the obliquity of the ecliptic, if
                            `date` is None.  But if `date` is given, then use that to determine the
                            epoch.  [default: 2000.]
        :param date:        If a date is given as a python datetime object, then return the
                            position in ecliptic coordinates with respect to the sun position at
                            that date.  If None, then return the true ecliptic coordiantes.
                            [default: None]

        :returns: the longitude and latitude as a tuple (lambda, beta), given as Angle instances.
        """
        # We are going to work in terms of the (x, y, z) projections.
        self._set_aux()

        # Get the obliquity of the ecliptic.
        if date is not None:
            epoch = date.year
        ep = util.ecliptic_obliquity(epoch)
        sin_ep, cos_ep = ep.sincos()

        # Coordinate transformation here, from celestial to ecliptic:
        x_ecl = self._x
        y_ecl = cos_ep*self._y + sin_ep*self._z
        z_ecl = -sin_ep*self._y + cos_ep*self._z

        beta = _Angle(math.asin(z_ecl))
        lam = _Angle(math.atan2(y_ecl, x_ecl))

        if date is not None:
            # Find the sun position in ecliptic coordinates on this date.  We have to convert to
            # Julian day in order to use our helper routine to find the Sun position in ecliptic
            # coordinates.
            lam_sun = util.sun_position_ecliptic(date)
            # Subtract it off, to get ecliptic coordinates relative to the sun.
            lam -= lam_sun

        return (lam.wrap(), beta)

    @staticmethod
    def from_ecliptic(lam, beta, epoch=2000., date=None):
        """Create a CelestialCoord from the given ecliptic coordinates

        :param lam:         The longitude in ecliptic coordinates (an Angle instance)
        :param beta:        The latitude in ecliptic coordinates (an Angle instance)
        :param epoch:       The epoch to be used for estimating the obliquity of the ecliptic, if
                            `date` is None.  But if `date` is given, then use that to determine the
                            epoch.  [default: 2000.]
        :param date:        If a date is given as a python datetime object, then return the
                            position in ecliptic coordinates with respect to the sun position at
                            that date.  If None, then return the true ecliptic coordiantes.
                            [default: None]

        :returns: the CelestialCoord corresponding to these ecliptic coordinates.
        """
        if date is not None:
            lam += util.sun_position_ecliptic(date)

        # Get the (x, y, z)_ecliptic from (lam, beta).
        sinbeta, cosbeta = beta.sincos()
        sinlam, coslam = lam.sincos()
        x_ecl = cosbeta*coslam
        y_ecl = cosbeta*sinlam
        z_ecl = sinbeta

        # Get the obliquity of the ecliptic.
        if date is not None:
            epoch = date.year
        ep = util.ecliptic_obliquity(epoch)

        # Transform to (x, y, z)_equatorial.
        sin_ep, cos_ep = ep.sincos()
        x_eq = x_ecl
        y_eq = cos_ep*y_ecl - sin_ep*z_ecl
        z_eq = sin_ep*y_ecl + cos_ep*z_ecl

        return CelestialCoord.from_xyz(x_eq, y_eq, z_eq)


    def __repr__(self): return 'coord.CelestialCoord(%r, %r)'%(self._ra,self._dec)
    def __str__(self): return 'coord.CelestialCoord(%s, %s)'%(self._ra,self._dec)
    def __hash__(self): return hash(repr(self))

    def __eq__(self, other):
        return (isinstance(other, CelestialCoord) and
                self.ra == other.ra and self.dec == other.dec)
    def __ne__(self, other): return not self.__eq__(other)

def _CelestialCoord(ra, dec):
    """
    Equivalent to CeletialCoord(ra,dec), but without the normal sanity checks.

    :param ra:       The right ascension.  Must be an Angle instance.
    :param dec:      The declination.  Must be an Angle instance.
    """
    ret = CelestialCoord.__new__(CelestialCoord)
    ret._ra = ra
    ret._dec = dec
    ret._x = None
    return ret



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Some utility functions used by the coord module.
"""

import numpy as np
import math
import datetime

from .angleunit import degrees
from .angle import Angle

def sun_position_ecliptic(date):
    """Helper routine to calculate the position of the sun in ecliptic coordinates given a
    python datetime object.

    It is most precise for dates between 1950-2050, and is based on

        http://en.wikipedia.org/wiki/Position_of_the_Sun#Ecliptic_coordinates

    :param date:    The date as either a datetime.datetime instance or a datetime.date instance.

    :returns the angular position of the sun along the ecliptic.
    """
    # We start by getting the number of days since Greenwich noon on 1 January 2000 (J2000).
    jd = date_to_julian_day(date)
    n = jd - 2451545.0
    L = 280.46*degrees + (0.9856474*degrees) * n
    g = 357.528*degrees + (0.9856003*degrees) * n
    lam = L + (1.915*degrees)*g.sin() + (0.020*degrees)*(2*g).sin()
    return lam

def date_to_julian_day(date):
    """Helper routine to return the Julian day for a given date.

    If `date` is a datetime.datetime instance, then it uses the full time info.
    If `date` is a datetime.date, then it does the calculation for noon of that day.

    :returns: the (possibly fractional) Julian day for the given date.
    """
    # From http://code-highlights.blogspot.com/2013/01/julian-date-in-python.html
    if not (isinstance(date, datetime.date) or isinstance(date, datetime.datetime)):
        raise ValueError("Date must be a python datetime object!")
    a = (14. - date.month)//12
    y = date.year + 4800 - a
    m = date.month + 12*a - 3
    retval = date.day + ((153*m + 2)//5) + 365*y + y//4 - y//100 + y//400 - 32045
    if isinstance(date, datetime.datetime):
        dayfrac = (date.hour + date.minute/60. + date.second/3600.)/24
        # The default is the value at noon, so we want to add 0 if dayfrac = 0.5
        dayfrac -= 0.5
        retval += dayfrac
    return retval

def ecliptic_obliquity(epoch):
    """Helper routine to return the obliquity of the ecliptic for a given date.

    :param epoch:   The epoch at which to calculate the obliquity.

    :returns the obliquity as an Angle instance
    """
    # We need to figure out the time in Julian centuries from J2000 for this epoch.
    t = (epoch - 2000.)/100.
    # Then we use the last (most recent) formula listed under
    # http://en.wikipedia.org/wiki/Ecliptic#Obliquity_of_the_ecliptic, from
    # JPL's 2010 calculations.
    ep = Angle.from_dms('23:26:21.406')
    ep -= Angle.from_dms('00:00:46.836769')*t
    ep -= Angle.from_dms('00:00:0.0001831')*(t**2)
    ep += Angle.from_dms('00:00:0.0020034')*(t**3)
    # There are even higher order terms, but they are probably not important for any reasonable
    # calculation someone would do with this package.
    return ep



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
import numpy as np

from .angleunit import AngleUnit, radians, degrees, hours, arcmin, arcsec

class Angle(object):
    """A class representing an Angle.  Angles are a value with an AngleUnit.

    **Initialization:**

        You typically create an Angle by multiplying a number by a coord.AngleUnit, for example:

            ..

            >>> pixel = 0.27 * arcsec
            >>> ra = 13.4 * hours
            >>> dec = -32 * degrees
            >>> from math import pi
            >>> theta = pi/2. * radians

        You can also initialize explicitly, taking a value and a unit:

            :meth:`coord.Angle.__init__`

            >>> unit = AngleUnit(math.pi / 100)  # gradians
            >>> phi = Angle(90, unit)

    **Built-in units:**

        There are five built-in AngleUnits which are always available for use:

            :coord.radians:   coord.AngleUnit(1.)
            :coord.degrees:   coord.AngleUnit(pi / 180.)
            :coord.hours:     coord.AngleUnit(pi / 12.)
            :coord.arcmin:    coord.AngleUnit(pi / 180. / 60.)
            :coord.arcsec:    coord.AngleUnit(pi / 180. / 3600.)

    **Attribute:**

        Since extracting the value in radians is extremely common, we have a read-only attribute
        to do this quickly:

            :rad:       The measure of the unit in radians.

        For example:

            ..

            >>> theta = 90 * degrees
            >>> print(theta.rad)
            1.57079632679

        It is equivalent to the more verbose:

            ..

            >>> x = theta / radians
            >>> print(x)
            1.57079632679

        but without actually requiring the floating point operation of dividing by 1.

    **Arithmetic:**

        Allowed arithmetic with Angles include the following.
        In the list below,

            - ``x`` is an arbitrary `float` value
            - ``unit1`` and ``unit2`` are arbitrary `coord.AngleUnit` instances
            - ``theta1`` and ``theta2`` are arbitrary `coord.Angle` instances

            >>> x = 37.8
            >>> unit1 = arcmin
            >>> unit2 = degrees

            >>> theta1 = x * unit1
            >>> theta2 = x * unit2
            >>> x2 = theta1 / unit2
            >>> theta = theta1 + theta2
            >>> theta = theta1 - theta2
            >>> theta = theta1 * x
            >>> theta = x * theta1
            >>> theta = theta1 / x
            >>> theta = -theta1
            >>> theta += theta1
            >>> theta -= theta1
            >>> theta *= x
            >>> theta /= x
            >>> x = unit1 / unit2   # equivalent to x = (1 * unit1) / unit2

        The above operations on NumPy arrays containing Angles are permitted as well.

    **Trigonometry:**

        There are convenience function for getting the sin, cos, and tan of an angle, along with
        one for getting sin and cos together, which should be more efficient than doing sin and
        cos separately:

            | :meth:`coord.Angle.sin`
            | :meth:`coord.Angle.cos`
            | :meth:`coord.Angle.tan`
            | :meth:`coord.Angle.sincos`

            >>> sint = theta.sin()  # equivalent to sint = math.sin(theta.rad)
            >>> cost = theta.cos()  # equivalent to cost = math.cos(theta.rad)
            >>> tant = theta.tan()  # equivalent to tant = math.tan(theta.rad)
            >>> sint, cost = theta.sincos()

        These functions mean that numpy trig functions will work on Angles or arrays of Angles:

            ..

            >>> sint = np.sin(theta)
            >>> cost = np.cos(theta)
            >>> tant = np.tan(theta)

    **Wrapping:**

        Depending on the context, theta = 2pi radians and theta = 0 radians may mean the same thing.
        If you want your angles to be wrapped to [-pi,pi) radians, you can do this by calling

            :meth:`coord.Angle.wrap`

            >>> theta = theta.wrap()

        This could be appropriate before testing for the equality of two angles for example, or
        calculating the difference between them.

        There is also an option to wrap into a different 2 pi range if so desired by specifying
        the center of the range.
    """
    def __init__(self, theta, unit=None):
        """
        :param theta:   The numerical value of the angle.
        :param unit:    The units theta is measured in.
        """
        # We also want to allow angle1 = Angle(angle2) as a copy, so check for that.
        if isinstance(theta,Angle):
            if unit is not None:
                raise TypeError("Cannot provide unit if theta is already an Angle instance")
            self._rad = theta._rad
        elif unit is None:
            raise TypeError("Must provide unit for Angle.__init__")
        elif not isinstance(unit, AngleUnit):
            raise TypeError("Invalid unit %s of type %s"%(unit,type(unit)))
        else:
            # Normal case
            self._rad = float(theta) * unit.value

    @property
    def rad(self):
        """Return the Angle in radians.

        Equivalent to angle / coord.radians
        """
        return self._rad

    def __neg__(self):
        return _Angle(-self._rad)

    def __add__(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Cannot add %s of type %s to an Angle"%(other,type(other)))
        return _Angle(self._rad + other._rad)

    def __sub__(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Cannot subtract %s of type %s from an Angle"%(other,type(other)))
        return _Angle(self._rad - other._rad)

    def __mul__(self, other):
        if other != float(other):
            raise TypeError("Cannot multiply Angle by %s of type %s"%(other,type(other)))
        return _Angle(self._rad * other)

    __rmul__ = __mul__

    def __div__(self, other):
        if isinstance(other, AngleUnit):
            return self._rad / other.value
        elif other == float(other):
            return _Angle(self._rad / other)
        else:
            raise TypeError("Cannot divide Angle by %s of type %s"%(other,type(other)))

    __truediv__ = __div__

    def wrap(self, center=None):
        """Wrap Angle to lie in the range [-pi, pi) radians (or other range  of 2pi radians)

        Depending on the context, theta = 2pi radians and theta = 0 radians are the same thing.
        If you want your angles to be wrapped to [-pi, pi) radians, you can do this as follows:

            ..

            >>> theta = Angle(700 * degrees)
            >>> theta = theta.wrap()
            >>> print(theta / degrees)
            -20.0

        This could be appropriate before testing for the equality of two angles for example, or
        calculating the difference between them.

        If you want to wrap to a different range than [-pi, pi), you can set the `center` argument
        to be the desired center of the the range.  e.g. for return values to fall in [0, 2pi),
        you could call

            ..

            >>> theta = theta.wrap(center=180. * degrees)
            >>> print(theta / degrees)
            340.0

        :param center:  The center point of the wrapped range. [default: 0 radians]

        :returns: the equivalent angle within the range [center-pi, center+pi)
        """
        if center is None: center = _Angle(0.)
        start = center._rad - math.pi
        offset = (self._rad - start) // (2.*math.pi)  # How many full cycles to subtract
        return _Angle(self._rad - offset * 2.*math.pi)

    def sin(self):
        """Return the sin of an Angle."""
        return math.sin(self._rad)

    def cos(self):
        """Return the cos of an Angle."""
        return math.cos(self._rad)

    def tan(self):
        """Return the tan of an Angle."""
        return math.tan(self._rad)

    def sincos(self):
        """Return both the sin and cos of an Angle as a numpy array [sint, cost].

        (On some systems, this may be slightly faster than doing each separately.)
        """
        # Note: This is admittedly pretty gratuitous.  We're writing Python after all, so the
        # difference between this and two trig calls is probably neglible given all the normal
        # Python overhead.  Mostly, I added this as an excuse to set up the C++ extension stuff
        # to make sure people could see how to add further C++-layer optimizations as needed.
        import coord
        sc = np.empty(2)
        coord._lib.coord_sincos(self._rad, coord._ffi.cast('double*', sc.ctypes.data))
        return sc

    def __str__(self):
        return str(self._rad) + ' radians'

    def __repr__(self):
        return 'coord.Angle(%r, coord.radians)'%self.rad

    def __eq__(self, other):
        return isinstance(other,Angle) and self.rad == other.rad

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(('coord.Angle', self._rad))

    @staticmethod
    def _make_dms_string(decimal, sep, prec, pad, plus_sign):
        # Account for the sign properly
        if decimal < 0:
            sign = '-'
            decimal = -decimal
        elif plus_sign:
            sign = '+'
        else:
            sign = ''

        # Figure out the 3 sep tokens
        sep1 = sep2 = ''
        sep3 = None
        if len(sep) == 1:
            sep1 = sep2 = sep
        elif len(sep) == 2:
            sep1, sep2 = sep
        elif len(sep) == 3:
            sep1, sep2, sep3 = sep

        # Round to nearest 1.e-8 seconds (or 10**-prec if given)
        round_prec = 8 if prec is None else prec
        digits = 10**round_prec

        decimal = int(3600 * digits * decimal + 0.5)

        d = decimal // (3600 * digits)
        decimal -= d * (3600 * digits)
        m = decimal // (60 * digits)
        decimal -= m * (60 * digits)
        s = decimal // digits
        decimal -= s * digits

        # Make the string
        if pad:
            d_str = '%02d'%d
            m_str = '%02d'%m
            s_str = '%02d'%s
        else:
            d_str = '%d'%d
            m_str = '%d'%m
            s_str = '%d'%s
        string = '%s%s%s%s%s%s.%0*d'%(sign,d_str,sep1,m_str,sep2,s_str,round_prec,decimal)
        if not prec:
            string = string.rstrip('0')
            string = string.rstrip('.')
        if sep3:
            string = string + sep3
        return string

    def hms(self, sep=":", prec=None, pad=True, plus_sign=False):
        """Return an HMS representation of the angle as a string: +-hh:mm:ss.decimal.

        An optional `sep` parameter can change the : to something else (e.g. a space or
        nothing at all).

        Note: the reverse process is effected by `from_hms`:

            ..

            >>> angle = -5.357 * hours
            >>> hms = angle.hms()
            >>> print(hms)
            -05:21:25.2
            >>> angle2 = Angle.from_hms(hms)
            >>> print(angle2 / hours)
            -5.357

        :param sep:         The token to put between the hh and mm and beteen mm and ss.  This may
                            also be a string of 2 or 3 items, e.g. 'hm' or 'hms'.  Or even a
                            tuple of strings such as ('hours ', 'minutes ', 'seconds').
                            [default: ':']
        :param prec:        The number of digits of precision after the decimal point.
                            [default: None]
        :param pad:         Whether to pad with a leading 0 if necessary to make h,m,s 2 digits.
                            [default: True]
        :param plus_sign:   Whether to use a plus sign for positive angles. [default: False]

        :returns: a string of the HMS representation of the angle.
        """
        if not len(sep) <= 3:
            raise ValueError("sep must be a string or tuple of length <= 3")
        if prec is not None and not prec >= 0:
            raise ValueError("prec must be >= 0")
        return self._make_dms_string(self/hours, sep, prec, pad, plus_sign)

    def dms(self, sep=":", prec=None, pad=True, plus_sign=False):
        """Return a DMS representation of the angle as a string: +-dd:mm:ss.decimal
        An optional `sep` parameter can change the : to something else (e.g. a space or
        nothing at all).

        Note: the reverse process is effected by `from_dms`:

            ..

            >>> angle = -(5 * degrees + 21 * arcmin + 25.2 * arcsec)
            >>> dms = angle.dms()
            >>> print(dms)
            -05:21:25.2
            >>> angle2 = Angle.from_dms(dms)
            >>> print(angle2 / degrees)
            -5.357

        :param sep:         The token to put between the hh and mm and beteen mm and ss.  This may
                            also be a string of 2 or 3 items, e.g. 'dm' or 'dms'.  Or even a
                            tuple of strings such as ('degrees ', 'minutes ', 'seconds').
                            [default: ':']
        :param prec:        The number of digits of precision after the decimal point.
                            [default: None]
        :param pad:         Whether to pad with a leading 0 if necessary to make h 2 digits.
                            [default: True]
        :param plus_sign:   Whether to use a plus sign for positive angles. [default: False]

        :returns: a string of the DMS representation of the angle.
        """
        if not len(sep) <= 3:
            raise ValueError("sep must be a string or tuple of length <= 3")
        if prec is not None and not prec >= 0:
            raise ValueError("prec must be >= 0")
        return self._make_dms_string(self/degrees, sep, prec, pad, plus_sign)

    @staticmethod
    def from_hms(str):
        """Convert a string of the form hh:mm:ss.decimal into an Angle.

        There may be an initial + or - (or neither), then two digits for the hours, two for the
        minutes, and two for the seconds.  Then there may be a decimal point followed by more
        digits.  There may be a colon separating hh, mm, and ss, or whitespace, or nothing at all.
        In fact, the code will ignore any non-digits between the hours, minutes, and seconds.

        Note: the reverse process is effected by Angle.hms():

            ..

            >>> angle = -5.357 * hours
            >>> hms = angle.hms()
            >>> print(hms)
            -05:21:25.2
            >>> angle2 = Angle.from_hms(hms)
            >>> print(angle2 / hours)
            -5.357

        :param str:     The string to parse.

        :returns: the corresponding Angle instance
        """
        return Angle._parse_dms(str) * hours

    @staticmethod
    def from_dms(str):
        """Convert a string of the form dd:mm:ss.decimal into an Angle.

        There may be an initial + or - (or neither), then two digits for the degrees, two for the
        minutes, and two for the seconds.  Then there may be a decimal point followed by more
        digits.  There may be a colon separating dd, mm, and ss, or whitespace, or nothing at all.
        In fact, the code will ignore any non-digits between the degrees, minutes, and seconds.

        Note: the reverse process is effected by Angle.dms():

            ..

            >>> angle = -(5 * degrees + 21 * arcmin + 25.2 * arcsec)
            >>> dms = angle.dms()
            >>> print(dms)
            -05:21:25.2
            >>> angle2 = Angle.from_dms(dms)
            >>> print(angle2 / degrees)
            -5.357

        :param str:     The string to parse.

        :returns: the corresponding Angle instance
        """
        return Angle._parse_dms(str) * degrees

    @staticmethod
    def _parse_dms(dms):
        """Convert a string of the form dd:mm:ss.decimal into decimal degrees.
        """
        import re
        tokens = tuple(filter(None, re.split('([\.\d]+)', dms.strip())))
        if len(tokens) <= 1:
            raise ValueError("string is not of the expected format")
        sign = 1
        try:
            dd = float(tokens[0])
        except ValueError:
            if tokens[0].strip() == '-':
                sign = -1
            tokens = tokens[1:]
            dd = float(tokens[0])
            if len(tokens) <= 1:
                raise ValueError("string is not of the expected format")
        if len(tokens) <= 2:
            return sign * dd
        mm = float(tokens[2])
        if len(tokens) <= 4:
            return sign * (dd + mm/60)
        if len(tokens) >= 7:
            raise ValueError("string is not of the expected format")
        ss = float(tokens[4])
        return sign * (dd + mm/60. + ss/3600.)

def _Angle(theta):
    """Equivalent to ``Angle(theta, coord.radians)``, but without the normal overhead (which isn't
    much to be honest, but this is nonetheless slightly quicker).

    :param theta:   The numerical value of the angle in radians.
    """
    ret = Angle.__new__(Angle)
    ret._rad = theta
    return ret




# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import numpy as np
import math
import coord

from helper_util import *

# We'll use these a lot, so just import them.
from numpy import pi, sin, cos, tan, arccos, sqrt
from coord import radians, degrees, hours, arcmin, arcsec

def check_aux(c):
    """A helper function to check that the aux values are set correctly (or get set correctly
    after calling set_aux()).
    """
    c._set_aux()
    np.testing.assert_almost_equal(c._sindec, c._dec.sin(), decimal=12)
    np.testing.assert_almost_equal(c._cosdec, c._dec.cos(), decimal=12)
    np.testing.assert_almost_equal(c._sinra, c._ra.sin(), decimal=12)
    np.testing.assert_almost_equal(c._cosra, c._ra.cos(), decimal=12)
    np.testing.assert_almost_equal(c._x, c._cosdec * c._cosra, decimal=12)
    np.testing.assert_almost_equal(c._y, c._cosdec * c._sinra, decimal=12)
    np.testing.assert_almost_equal(c._z, c._sindec, decimal=12)


@timer
def test_init():
    """Basic tests of CelestialCoord construction.
    """
    c1 = coord.CelestialCoord(0. * radians, 0. * radians)
    np.testing.assert_almost_equal(c1.ra.rad, 0., decimal=12)
    np.testing.assert_almost_equal(c1.dec.rad, 0., decimal=12)

    # These next three are the same place
    c2 = coord.CelestialCoord(11. * hours, -37. * degrees)
    np.testing.assert_almost_equal(c2.ra / hours, 11., decimal=12)
    np.testing.assert_almost_equal(c2.dec / degrees, -37., decimal=12)

    c3 = coord.CelestialCoord(35. * hours, -37. * degrees)
    np.testing.assert_almost_equal(c3.ra / hours, 35., decimal=12)
    np.testing.assert_almost_equal(c3.dec / degrees, -37., decimal=12)
    np.testing.assert_almost_equal(c3.normal().ra / hours, 11., decimal=12)
    np.testing.assert_almost_equal(c3.normal().dec / degrees, -37., decimal=12)

    c4 = coord.CelestialCoord(-13. * hours, -37. * degrees)
    np.testing.assert_almost_equal(c4.ra / hours, -13., decimal=12)
    np.testing.assert_almost_equal(c4.dec / degrees, -37., decimal=12)
    np.testing.assert_almost_equal(c4.normal().ra / hours, 11., decimal=12)
    np.testing.assert_almost_equal(c4.normal().dec / degrees, -37., decimal=12)

    # copy constructor
    c5 = coord.CelestialCoord(c2)
    np.testing.assert_almost_equal(c5.ra / hours, 11., decimal=12)
    np.testing.assert_almost_equal(c5.dec / degrees, -37., decimal=12)

    # We'll test distance later, but for check that these last 3 have distances = 0.
    np.testing.assert_almost_equal(c2.distanceTo(c3).rad, 0., decimal=12)
    np.testing.assert_almost_equal(c2.distanceTo(c4).rad, 0., decimal=12)

    # Check that the set_aux function properly sets some auxiliary values
    check_aux(c1)
    check_aux(c2)
    check_aux(c3)
    check_aux(c4)

    # Check picklability
    do_pickle(c1)
    do_pickle(c2)
    do_pickle(c3)
    do_pickle(c4)

    # Check some invalid values
    np.testing.assert_raises(TypeError, coord.CelestialCoord, 11 * hours)
    np.testing.assert_raises(TypeError, coord.CelestialCoord, 11 * hours, -37)
    np.testing.assert_raises(TypeError, coord.CelestialCoord, 11, -37 * degrees)
    np.testing.assert_raises(ValueError, coord.CelestialCoord, 11. * hours, 99 * degrees)


@timer
def test_eq():
    """Check that equal coords are equal, but unequal ones are not.
    """
    c1 = coord.CelestialCoord(11. * hours, -37. * degrees)
    c2 = coord.CelestialCoord(165. * degrees, -37. * degrees)
    assert c2 == c1

    # These should all test as unequal.  Note some non-Angles in the list.
    diff_list = [ c1,
                  coord.CelestialCoord(0 * radians, 0 * radians),
                  coord.CelestialCoord(13. * hours, 37. * degrees),
                  coord.CelestialCoord(11. * hours, 37. * degrees),
                  pi, coord.CelestialCoord, None ]
    all_obj_diff(diff_list)


@timer
def test_distance():
    """Test calculations of distances on the sphere.
    """
    # First, let's test some distances that are easy to figure out without any spherical trig.
    eq1 = coord.CelestialCoord(0 * radians, 0 * radians)  # point on the equator
    eq2 = coord.CelestialCoord(1 * radians, 0 * radians)  # 1 radian along equator
    eq3 = coord.CelestialCoord(pi * radians, 0 * radians)  # antipode of eq1
    north_pole = coord.CelestialCoord(0 * radians, pi/2 * radians)  # north pole
    south_pole = coord.CelestialCoord(0 * radians, -pi/2 * radians) # south pole

    np.testing.assert_almost_equal(eq1.distanceTo(eq2).rad, 1, decimal=12)
    np.testing.assert_almost_equal(eq2.distanceTo(eq1).rad, 1, decimal=12)
    np.testing.assert_almost_equal(eq1.distanceTo(eq3).rad, pi, decimal=12)
    np.testing.assert_almost_equal(eq2.distanceTo(eq3).rad, pi-1, decimal=12)

    np.testing.assert_almost_equal(north_pole.distanceTo(south_pole).rad, pi, decimal=12)

    np.testing.assert_almost_equal(eq1.distanceTo(north_pole).rad, pi/2, decimal=12)
    np.testing.assert_almost_equal(eq2.distanceTo(north_pole).rad, pi/2, decimal=12)
    np.testing.assert_almost_equal(eq3.distanceTo(north_pole).rad, pi/2, decimal=12)
    np.testing.assert_almost_equal(eq1.distanceTo(south_pole).rad, pi/2, decimal=12)
    np.testing.assert_almost_equal(eq2.distanceTo(south_pole).rad, pi/2, decimal=12)
    np.testing.assert_almost_equal(eq3.distanceTo(south_pole).rad, pi/2, decimal=12)

    # Some random point
    c1 = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    # Same meridian
    c2 = coord.CelestialCoord(0.234 * radians, -1.093 * radians)
    # Antipode
    c3 = coord.CelestialCoord((pi + 0.234) * radians, -0.342 * radians)
    # Different point on opposide meridian
    c4 = coord.CelestialCoord((pi + 0.234) * radians, 0.832 * radians)

    for c, d in zip( (c1,c2,c3,c4), (0., 1.435, pi, pi-1.174) ):
        np.testing.assert_almost_equal(c1.distanceTo(c).rad, d, decimal=12)

    # Now some that require spherical trig calculations.
    # Importantly, this uses the more straightforward spherical trig formula, the cosine rule.
    # The CelestialCoord class uses a different formula that is more stable for very small
    # distances, which are typical in the correlation function calculation.
    # Some other random point:
    c5 = coord.CelestialCoord(1.832 * radians, -0.723 * radians)
    # The standard formula is:
    # cos(d) = sin(dec1) sin(dec2) + cos(dec1) cos(dec2) cos(delta ra)
    d = arccos(sin(c1.dec) * sin(c5.dec) + cos(c1.dec) * cos(c5.dec) * cos(c1.ra-c5.ra))
    np.testing.assert_almost_equal(c1.distanceTo(c5).rad, d, decimal=12)

    # Tiny displacements should have dsq = (dra^2 cos^2 dec) + (ddec^2)
    c6 = coord.CelestialCoord(c1.ra + 1.7e-9 * radians, c1.dec)
    c7 = coord.CelestialCoord(c1.ra, c1.dec + 1.9e-9 * radians)
    c8 = coord.CelestialCoord(c1.ra + 2.3e-9 * radians, c1.dec + 1.2e-9 * radians)

    # Note that the standard formula gets these wrong.  d is 0.0
    d = arccos(sin(c1.dec) * sin(c6.dec) + cos(c1.dec) * cos(c6.dec) * cos(c1.ra-c6.ra))
    print('d(c6) = ',1.7e-9 * cos(c1.dec), c1.distanceTo(c6), d)
    d = arccos(sin(c1.dec) * sin(c7.dec) + cos(c1.dec) * cos(c7.dec) * cos(c1.ra-c7.ra))
    print('d(c7) = ',1.9e-9, c1.distanceTo(c7), d)
    d = arccos(sin(c1.dec) * sin(c8.dec) + cos(c1.dec) * cos(c8.dec) * cos(c1.ra-c8.ra))
    true_d = sqrt( (2.3e-9 * cos(c1.dec))**2 + 1.2e-9**2)
    print('d(c7) = ',true_d, c1.distanceTo(c8), d)
    np.testing.assert_allclose(c1.distanceTo(c6).rad, 1.7e-9 * cos(c1.dec), rtol=1.e-7)
    np.testing.assert_allclose(c1.distanceTo(c7).rad, 1.9e-9, rtol=1.e-7)
    np.testing.assert_allclose(c1.distanceTo(c8).rad, true_d, rtol=1.e-7)

    # Near antipodes, the formula we usually use becomes somewhat inaccurate.
    # Check a variety of antipodes.
    for c in [c1, c2, c3, c4, c5, c6, c7, c8, eq1, eq2, eq3, north_pole, south_pole]:
        antipode = coord.CelestialCoord(c.ra + pi*radians, -c.dec)
        np.testing.assert_almost_equal(c.distanceTo(antipode).rad, pi, decimal=12)
        np.testing.assert_almost_equal(antipode.distanceTo(c).rad, pi, decimal=12)

    # Also some near, but not quite antipodes
    # Note: this range crosses the point where the formula in distanceTo changes.
    for delta in range(500):
        eq4 = coord.CelestialCoord(delta * arcmin, 0 * radians)
        np.testing.assert_allclose(pi - eq3.distanceTo(eq4).rad, eq4.ra.rad, rtol=1.e-7)


@timer
def test_xyz():
    """Test get_xyz and from_xyz functions
    """
    # Set up test coordinates
    tra = 12.0 * coord.hours
    tdec = 31.0 * coord.degrees
    tx = tdec.cos() * tra.cos()
    ty = tdec.cos() * tra.sin()
    tz = tdec.sin()
    
    # Check get_xyz by comparing to true values
    c1 = coord.CelestialCoord(ra=tra, dec=tdec)
    np.testing.assert_almost_equal(c1.get_xyz(), (tx, ty, tz), decimal=12)
    
    # Check ra and dec when setting coordinates with from_xyz
    c2 = coord.CelestialCoord.from_xyz(tx, ty, tz)
    np.testing.assert_almost_equal(c2.ra.rad, tra.rad, decimal=12,
                                   err_msg="RA from (x,y,z) not equal to true RA")
    np.testing.assert_almost_equal(c2.dec.rad, tdec.rad, decimal=12,
                                   err_msg="DEC from (x,y,z) not equal to true DEC")

    # Check that aux values are set up correctly.  The from_xyz function sets these itself,
    # so this is a non-trivial test.
    check_aux(c2)

    # Check that the x, y, z is consistent
    np.testing.assert_almost_equal(c2.get_xyz(), (tx, ty, tz), decimal=12,
                                   err_msg="Different (x, y, z) coordinates output than input.")

    # Check some special coordinates in case anything is funny at the poles or equator.
    eq1 = coord.CelestialCoord(0. * radians, 0. * radians)     # points on the equator
    eq2 = coord.CelestialCoord(0.5*pi * radians, 0. * radians)
    eq3 = coord.CelestialCoord(pi * radians, 0. * radians)
    eq4 = coord.CelestialCoord(1.5*pi * radians, 0. * radians)
    north_pole = coord.CelestialCoord(0. * radians, pi/2. * radians)  # north pole
    south_pole = coord.CelestialCoord(0. * radians, -pi/2. * radians) # south pole

    for c in [c1, eq1, eq2, eq3, eq4, north_pole, south_pole]:
        x, y, z = c.get_xyz()
        np.testing.assert_almost_equal(x, cos(c.dec) * cos(c.ra), decimal=12)
        np.testing.assert_almost_equal(y, cos(c.dec) * sin(c.ra), decimal=12)
        np.testing.assert_almost_equal(z, sin(c.dec), decimal=12)

        c2 = coord.CelestialCoord.from_xyz(x,y,z)
        np.testing.assert_almost_equal(c2.ra.rad, c.ra.rad, decimal=12)
        np.testing.assert_almost_equal(c2.dec.rad, c.dec.rad, decimal=12)
        check_aux(c2)

        # Check that it works if x,y,z are scaled arbitrarily
        c3 = coord.CelestialCoord.from_xyz(x*17,y*17,z*17)
        np.testing.assert_almost_equal(c3.ra.rad, c.ra.rad, decimal=12)
        np.testing.assert_almost_equal(c3.dec.rad, c.dec.rad, decimal=12)
        check_aux(c3)

        c4 = coord.CelestialCoord.from_xyz(x*1.e-9,y*1.e-9,z*1.e-9)
        np.testing.assert_almost_equal(c4.ra.rad, c.ra.rad, decimal=12)
        np.testing.assert_almost_equal(c4.dec.rad, c.dec.rad, decimal=12)
        check_aux(c4)

    # constructing from x,y,z = 0,0,0 is undefined.
    np.testing.assert_raises(ValueError, coord.CelestialCoord.from_xyz, 0., 0., 0.)


@timer
def test_xyz_array():
    """Test xyz_to_radec and radec_to_xyz
    """
    # Set up a list of ra, dec values to check.
    # Note: These include the same special locations used above in test_xyz.

    ra_ar = np.array([0.123, 0.55, 0., 0.5*pi, pi, 1.5*pi, 0., 0.])
    dec_ar = np.array([-0.5, 0.88, 0., 0., 0., 0., pi/2., -pi/2])

    x_ar = np.cos(dec_ar) * np.cos(ra_ar)
    y_ar = np.cos(dec_ar) * np.sin(ra_ar)
    z_ar = np.sin(dec_ar)
    x_ar[-2:] = 0.  # Make sure these are precisely 0 to test edge case of exactly n or s pole
    y_ar[-2:] = 0.

    # Check converting singly:
    for (x,y,z,ra,dec) in zip(x_ar, y_ar, z_ar, ra_ar, dec_ar):
        x1, y1, z1 = coord.CelestialCoord(ra*radians, dec*radians).get_xyz()
        assert np.isclose(x1, x)
        assert np.isclose(y1, y)
        assert np.isclose(z1, z)

        x2, y2, z2 = coord.CelestialCoord.radec_to_xyz(ra, dec)
        assert np.isclose(x2, x)
        assert np.isclose(y2, y)
        assert np.isclose(z2, z)

        x3, y3, z3 = coord.CelestialCoord.radec_to_xyz(ra, dec, r=17)
        assert np.isclose(x3, x*17)
        assert np.isclose(y3, y*17)
        assert np.isclose(z3, z*17)

        c = coord.CelestialCoord.from_xyz(x, y, z)
        ra1, dec1 = c.ra.rad, c.dec.rad
        assert np.isclose(ra1, ra)
        assert np.isclose(dec1, dec)

        ra2, dec2 = coord.CelestialCoord.xyz_to_radec(x, y, z)
        if ra2 < 0.: ra2 += 2.*pi
        assert np.isclose(ra2, ra)
        assert np.isclose(dec2, dec)

        ra3, dec3, r3 = coord.CelestialCoord.xyz_to_radec(x, y, z, return_r=True)
        if ra3 < 0.: ra3 += 2.*pi
        assert np.isclose(ra3, ra)
        assert np.isclose(dec3, dec)
        assert np.isclose(r3, (x**2+y**2+z**2)**0.5)

    # Now check converting them all at once
    x4, y4, z4 = coord.CelestialCoord.radec_to_xyz(ra_ar, dec_ar)
    np.testing.assert_allclose(x4, x_ar, rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(y4, y_ar, rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(z4, z_ar, rtol=1.e-8, atol=1.e-12)

    ra4, dec4 = coord.CelestialCoord.xyz_to_radec(x_ar, y_ar, z_ar)
    # check sin, cor, rather than angle, so wrap doesn't matter.
    np.testing.assert_allclose(np.cos(ra4), np.cos(ra_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(np.sin(ra4), np.sin(ra_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(np.cos(dec4), np.cos(dec_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(np.sin(dec4), np.sin(dec_ar), rtol=1.e-8, atol=1.e-12)

    # With r values
    r_ar = np.linspace(1,100, num=len(ra_ar))
    x5, y5, z5 = coord.CelestialCoord.radec_to_xyz(ra_ar, dec_ar, r_ar)
    np.testing.assert_allclose(x5, x_ar * r_ar, rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(y5, y_ar * r_ar, rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(z5, z_ar * r_ar, rtol=1.e-8, atol=1.e-12)

    ra5, dec5, r5 = coord.CelestialCoord.xyz_to_radec(x5, y5, z5, return_r=True)
    np.testing.assert_allclose(np.cos(ra5), np.cos(ra_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(np.sin(ra5), np.sin(ra_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(np.cos(dec5), np.cos(dec_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(np.sin(dec5), np.sin(dec_ar), rtol=1.e-8, atol=1.e-12)
    np.testing.assert_allclose(r5, r_ar, rtol=1.e-8, atol=1.e-12)


@timer
def test_angleBetween():
    """Test calculations of angles between positions on the sphere.
    """
    # Again, let's start with some answers we can get by inspection.
    eq1 = coord.CelestialCoord(0. * radians, 0. * radians)  # point on the equator
    eq2 = coord.CelestialCoord(1. * radians, 0. * radians)  # 1 radian along equator
    eq3 = coord.CelestialCoord(pi * radians, 0. * radians) # antipode of eq1
    north_pole = coord.CelestialCoord(0. * radians, pi/2. * radians)  # north pole
    south_pole = coord.CelestialCoord(0. * radians, -pi/2. * radians) # south pole

    np.testing.assert_almost_equal(north_pole.angleBetween(eq1,eq2).rad, -1., decimal=12)
    np.testing.assert_almost_equal(north_pole.angleBetween(eq2,eq1).rad, 1., decimal=12)
    np.testing.assert_almost_equal(north_pole.angleBetween(eq2,eq3).rad, 1.-pi, decimal=12)
    np.testing.assert_almost_equal(north_pole.angleBetween(eq3,eq2).rad, pi-1., decimal=12)
    np.testing.assert_almost_equal(south_pole.angleBetween(eq1,eq2).rad, 1., decimal=12)
    np.testing.assert_almost_equal(south_pole.angleBetween(eq2,eq1).rad, -1., decimal=12)
    np.testing.assert_almost_equal(south_pole.angleBetween(eq2,eq3).rad, pi-1., decimal=12)
    np.testing.assert_almost_equal(south_pole.angleBetween(eq3,eq2).rad, 1.-pi, decimal=12)

    np.testing.assert_almost_equal(eq1.angleBetween(north_pole,eq2).rad, pi/2., decimal=12)
    np.testing.assert_almost_equal(eq2.angleBetween(north_pole,eq1).rad, -pi/2., decimal=12)

    np.testing.assert_almost_equal(north_pole.area(eq1,eq2), 1., decimal=12)
    np.testing.assert_almost_equal(north_pole.area(eq2,eq1), 1., decimal=12)
    np.testing.assert_almost_equal(south_pole.area(eq1,eq2), 1., decimal=12)
    np.testing.assert_almost_equal(south_pole.area(eq2,eq1), 1., decimal=12)

    # For arbitrary points, we can check that the spherical triangle satisfies
    # the spherical trig laws.
    cA = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    cB = coord.CelestialCoord(-0.193 * radians, 0.882 * radians)
    cC = coord.CelestialCoord(0.721 * radians, -0.561 * radians)

    a = cB.distanceTo(cC)
    b = cC.distanceTo(cA)
    c = cA.distanceTo(cB)
    A = cA.angleBetween(cB,cC)
    B = cB.angleBetween(cC,cA)
    C = cC.angleBetween(cA,cB)
    E = abs(A.rad)+abs(B.rad)+abs(C.rad)-pi
    s = (a+b+c)/2.

    # Law of cosines:
    np.testing.assert_almost_equal(cos(c), cos(a)*cos(b) + sin(a)*sin(b)*cos(C), decimal=12)
    np.testing.assert_almost_equal(cos(a), cos(b)*cos(c) + sin(b)*sin(c)*cos(A), decimal=12)
    np.testing.assert_almost_equal(cos(b), cos(c)*cos(a) + sin(c)*sin(a)*cos(B), decimal=12)

    # Law of sines:
    np.testing.assert_almost_equal(sin(A) * sin(b), sin(B) * sin(a), decimal=12)
    np.testing.assert_almost_equal(sin(B) * sin(c), sin(C) * sin(b), decimal=12)
    np.testing.assert_almost_equal(sin(C) * sin(a), sin(A) * sin(c), decimal=12)

    # Alternate law of cosines:
    np.testing.assert_almost_equal(cos(C), -cos(A)*cos(B) + sin(A)*sin(B)*cos(c), decimal=12)
    np.testing.assert_almost_equal(cos(A), -cos(B)*cos(C) + sin(B)*sin(C)*cos(a), decimal=12)
    np.testing.assert_almost_equal(cos(B), -cos(C)*cos(A) + sin(C)*sin(A)*cos(b), decimal=12)

    # Spherical excess:
    np.testing.assert_almost_equal(cA.area(cB,cC), E, decimal=12)
    np.testing.assert_almost_equal(cA.area(cC,cB), E, decimal=12)
    np.testing.assert_almost_equal(cB.area(cA,cC), E, decimal=12)
    np.testing.assert_almost_equal(cB.area(cC,cA), E, decimal=12)
    np.testing.assert_almost_equal(cC.area(cB,cA), E, decimal=12)
    np.testing.assert_almost_equal(cC.area(cA,cB), E, decimal=12)

    # L'Huilier's formula for spherical excess:
    np.testing.assert_almost_equal(
            tan(E/4)**2,
            (s/2).tan() * ((s-a)/2).tan() * ((s-b)/2).tan() * ((s-c)/2).tan(), decimal=12)


@timer
def test_gnomonic_projection():
    """Test the gnomonic projection.
    """
    # Test that a small triangle has the correct properties.
    # Gnomonic projections turn great circles into straight lines.
    center = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    cA = coord.CelestialCoord(-0.193 * radians, 0.882 * radians)
    cB = coord.CelestialCoord((-0.193 + 1.7e-8) * radians,
                               (0.882 + 1.2e-8) * radians)
    cC = coord.CelestialCoord((-0.193 - 2.4e-8) * radians,
                               (0.882 + 3.1e-8) * radians)

    a = cB.distanceTo(cC).rad
    b = cC.distanceTo(cA).rad
    c = cA.distanceTo(cB).rad
    A = cA.angleBetween(cB,cC).rad
    B = cB.angleBetween(cC,cA).rad
    C = cC.angleBetween(cA,cB).rad
    E = cA.area(cB,cC)

    uA, vA = center.project(cA, projection='gnomonic')
    uB, vB = center.project(cB, projection='gnomonic')
    uC, vC = center.project(cC, projection='gnomonic')

    # Check that project_rad does the same thing
    uA2, vA2 = center.project_rad(cA.ra.rad, cA.dec.rad, projection='gnomonic')
    np.testing.assert_allclose([uA2,vA2], [uA.rad,vA.rad], rtol=1.e-8,
                               err_msg="project_rad not equivalent")

    # Check the deprojection
    cA2 = center.deproject(uA, vA, projection='gnomonic')
    np.testing.assert_allclose(cA2.rad, cA.rad, err_msg="deproject didn't return to orig")
    cA3 = center.deproject_rad(uA.rad, vA.rad, projection='gnomonic')
    np.testing.assert_allclose(cA3, cA.rad, err_msg="deproject_rad not equivalent")
    check_aux(cA2)

    # The angles are not preserved
    a = sqrt( (uB.rad-uC.rad)**2 + (vB.rad-vC.rad)**2 )
    b = sqrt( (uC.rad-uA.rad)**2 + (vC.rad-vA.rad)**2 )
    c = sqrt( (uA.rad-uB.rad)**2 + (vA.rad-vB.rad)**2 )
    cosA = ((uB.rad-uA.rad)*(uC.rad-uA.rad) + (vB.rad-vA.rad)*(vC.rad-vA.rad)) / (b*c)
    cosB = ((uC.rad-uB.rad)*(uA.rad-uB.rad) + (vC.rad-vB.rad)*(vA.rad-vB.rad)) / (c*a)
    cosC = ((uA.rad-uC.rad)*(uB.rad-uC.rad) + (vA.rad-vC.rad)*(vB.rad-vC.rad)) / (a*b)

    print('gnomonic cosA = ',cosA,cos(A))
    print('gnomonic cosB = ',cosB,cos(B))
    print('gnomonic cosC = ',cosC,cos(C))

    # The area is not preserved
    area = 0.5 * abs( (uB.rad-uA.rad) * (vC.rad-vA.rad) - (uC.rad-uA.rad) * (vB.rad-vA.rad) )
    print('gnomonic area = ',area,E)

    # The deproject jacobian should tell us how the area changes
    dudx, dudy, dvdx, dvdy = center.jac_deproject(uA, vA, 'gnomonic').ravel()
    jac_area = abs(dudx*dvdy - dudy*dvdx)
    np.testing.assert_allclose(jac_area, E/area, err_msg='jac_deproject gave wrong area')

    dudx, dudy, dvdx, dvdy = center.jac_deproject_rad(uA.rad, vA.rad, 'gnomonic').ravel()
    np.testing.assert_allclose(jac_area, abs(dudx*dvdy - dudy*dvdx),
                               err_msg='jac_deproject_rad not equivalent')

    # center projects to 0,0 with unit area
    u, v = center.project(center, 'gnomonic')
    np.testing.assert_allclose([u.rad, v.rad], 0., err_msg='center did not project to (0,0)')
    c2 = center.deproject(u,v, 'gnomonic')
    np.testing.assert_allclose(c2.rad, center.rad, err_msg='(0,0) did not deproject to center')
    np.testing.assert_allclose(np.linalg.det(center.jac_deproject(u, v, 'gnomonic')), 1.,
                               err_msg='determinant of jac_deproject matrix != 1 at center')
    check_aux(c2)

    # Finally, check the claim that great circles turn into straight lines.
    # Meridians are great circles.  Also symmetric points across the equator.
    center = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    gc_list = [ [ (0.1, -0.5), (0.1, -0.3), (0.1, 0.1), (0.1, 0.4) ],
                [ (0.3, -0.5), (0.3, -0.3), (0.3, 0.1), (0.3, 0.4) ],
                [ (0.5, -0.5), (0.5, -0.3), (0.5, 0.1), (0.5, 0.4) ],
                [ (0.7, -0.5), (0.7, -0.3), (0.7, 0.1), (0.7, 0.4) ],
                [ (-0.1, -0.4), (0.1, 0.0), (0.3, 0.4) ],
                [ (0.3, -0.2), (0.4, 0.0), (0.5, 0.2) ],
                [ (0.3, -0.1), (0.0, 0.0), (-0.3, 0.1) ] ]
    for gc in gc_list:
        print('gc = ',gc)
        # The projected points for a single great circle
        proj = [center.project(coord.CelestialCoord(c[0] * radians, c[1] * radians)) for c in gc]
        # Calculate the slope from the first and last points
        u0, v0 = proj[0]
        un, vn = proj[-1]
        m0 = (vn.rad - v0.rad) / (un.rad - u0.rad)
        print('m0 = ',m0)
        for (u,v) in proj[1:-1]:
            m1 = (v.rad - v0.rad) / (u.rad - u0.rad)
            print('u,v = ',u,v,', m1 = ',m1)
            np.testing.assert_almost_equal(m1, m0, decimal=12)

    # gnomonic is the default.  Make sure that works right.
    uA3, vA3 = center.project(cA)
    uB3, vB3 = center.project(cB)
    uC3, vC3 = center.project(cC)
    np.testing.assert_array_almost_equal([uA3.rad, vA3.rad], [uA.rad, vA.rad], decimal=12)
    np.testing.assert_array_almost_equal([uB3.rad, vB3.rad], [uB.rad, vB.rad], decimal=12)
    np.testing.assert_array_almost_equal([uB3.rad, vB3.rad], [uB.rad, vB.rad], decimal=12)

    # And make sure that invalid projection strings raise exceptions
    np.testing.assert_raises(ValueError, center.project, cA, 'TAN')
    np.testing.assert_raises(ValueError, center.project, cA, projection=3)
    np.testing.assert_raises(ValueError, center.project, 3, 4)
    np.testing.assert_raises(ValueError, center.project, u, v)
    np.testing.assert_raises(ValueError, center.project_rad, 3, 4, 'TAN')
    np.testing.assert_raises(ValueError, center.project_rad, 3, 4, projection=3)
    np.testing.assert_raises(TypeError, center.project_rad, cA)
    np.testing.assert_raises(ValueError, center.deproject, u, v, 'TAN')
    np.testing.assert_raises(ValueError, center.deproject, u, v, projection=3)
    np.testing.assert_raises(TypeError, center.deproject, 3, 4)
    np.testing.assert_raises(TypeError, center.deproject, cA)
    np.testing.assert_raises(ValueError, center.deproject_rad, 3, 4, 'TAN')
    np.testing.assert_raises(ValueError, center.deproject_rad, 3, 4, projection=3)
    np.testing.assert_raises(TypeError, center.deproject_rad, u, v)
    np.testing.assert_raises(TypeError, center.deproject_rad, cA)
    np.testing.assert_raises(ValueError, center.jac_deproject, u, v, 'TAN')
    np.testing.assert_raises(ValueError, center.jac_deproject, u, v, projection=3)
    np.testing.assert_raises(AttributeError, center.jac_deproject, 3, 4)
    np.testing.assert_raises(TypeError, center.jac_deproject, cA)
    np.testing.assert_raises(ValueError, center.jac_deproject_rad, 3, 4, 'TAN')
    np.testing.assert_raises(ValueError, center.jac_deproject_rad, 3, 4, projection=3)
    np.testing.assert_raises(TypeError, center.jac_deproject_rad, u, v)
    np.testing.assert_raises(TypeError, center.jac_deproject_rad, cA)


@timer
def test_stereographic_projection():
    """Test the stereographic projection.
    """
    # Test that a small triangle has the correct properties.
    # Stereographic projections preserve angles, but not area.
    center = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    cA = coord.CelestialCoord(-0.193 * radians, 0.882 * radians)
    cB = coord.CelestialCoord((-0.193 + 1.7e-8) * radians,
                               (0.882 + 1.2e-8) * radians)
    cC = coord.CelestialCoord((-0.193 - 2.4e-8) * radians,
                               (0.882 + 3.1e-8) * radians)

    a = cB.distanceTo(cC).rad
    b = cC.distanceTo(cA).rad
    c = cA.distanceTo(cB).rad
    A = cA.angleBetween(cB,cC).rad
    B = cB.angleBetween(cC,cA).rad
    C = cC.angleBetween(cA,cB).rad
    E = cA.area(cB,cC)

    uA, vA = center.project(cA, projection='stereographic')
    uB, vB = center.project(cB, projection='stereographic')
    uC, vC = center.project(cC, projection='stereographic')

    # The easiest way to compute the angles is from the dot products:
    # a.b = ab cos(C)
    a = sqrt( (uB.rad-uC.rad)**2 + (vB.rad-vC.rad)**2 )
    b = sqrt( (uC.rad-uA.rad)**2 + (vC.rad-vA.rad)**2 )
    c = sqrt( (uA.rad-uB.rad)**2 + (vA.rad-vB.rad)**2 )
    cosA = ((uB.rad-uA.rad)*(uC.rad-uA.rad) + (vB.rad-vA.rad)*(vC.rad-vA.rad)) / (b*c)
    cosB = ((uC.rad-uB.rad)*(uA.rad-uB.rad) + (vC.rad-vB.rad)*(vA.rad-vB.rad)) / (c*a)
    cosC = ((uA.rad-uC.rad)*(uB.rad-uC.rad) + (vA.rad-vC.rad)*(vB.rad-vC.rad)) / (a*b)

    print('stereographic cosA = ',cosA,cos(A))
    print('stereographic cosB = ',cosB,cos(B))
    print('stereographic cosC = ',cosC,cos(C))
    np.testing.assert_allclose(cosA, cos(A))
    np.testing.assert_allclose(cosB, cos(B))
    np.testing.assert_allclose(cosC, cos(C))

    # Check that project_rad does the same thing
    uA2, vA2 = center.project_rad(cA.ra.rad, cA.dec.rad, projection='stereographic')
    np.testing.assert_allclose([uA2,vA2], [uA.rad,vA.rad], rtol=1.e-8,
                               err_msg="project_rad not equivalent")

    # Check the deprojection
    cA2 = center.deproject(uA, vA, projection='stereographic')
    np.testing.assert_allclose(cA2.rad, cA.rad, err_msg="deproject didn't return to orig")
    cA3 = center.deproject_rad(uA.rad, vA.rad, projection='stereographic')
    np.testing.assert_allclose(cA3, cA.rad, err_msg="deproject_rad not equivalent")
    check_aux(cA2)

    # The area is not preserved
    area = 0.5 * abs( (uB.rad-uA.rad) * (vC.rad-vA.rad) - (uC.rad-uA.rad) * (vB.rad-vA.rad) )
    print('stereographic area = ',area,E)

    # The deproject jacobian should tell us how the area changes
    dudx, dudy, dvdx, dvdy = center.jac_deproject(uA, vA, 'stereographic').ravel()
    jac_area = abs(dudx*dvdy - dudy*dvdx)
    np.testing.assert_allclose(jac_area, E/area, err_msg='jac_deproject gave wrong area')

    dudx, dudy, dvdx, dvdy = center.jac_deproject_rad(uA.rad, vA.rad, 'stereographic').ravel()
    np.testing.assert_allclose(jac_area, abs(dudx*dvdy - dudy*dvdx),
                               err_msg='jac_deproject_rad not equivalent')

    # center projects to 0,0 with unit area
    u, v = center.project(center, 'stereographic')
    np.testing.assert_allclose([u.rad, v.rad], 0., err_msg='center did not project to (0,0)')
    c2 = center.deproject(u,v, 'stereographic')
    np.testing.assert_allclose(c2.rad, center.rad, err_msg='(0,0) did not deproject to center')
    np.testing.assert_allclose(np.linalg.det(center.jac_deproject(u, v, 'stereographic')), 1.,
                               err_msg='determinant of jac_deproject matrix != 1 at center')
    check_aux(c2)


@timer
def test_lambert_projection():
    """Test the lambert projection.
    """
    # Test that a small triangle has the correct properties.
    # Lambert projections preserve area, but not angles.
    center = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    cA = coord.CelestialCoord(-0.193 * radians, 0.882 * radians)
    cB = coord.CelestialCoord((-0.193 + 1.7e-8) * radians,
                               (0.882 + 1.2e-8) * radians)
    cC = coord.CelestialCoord((-0.193 - 2.4e-8) * radians,
                               (0.882 + 3.1e-8) * radians)

    a = cB.distanceTo(cC).rad
    b = cC.distanceTo(cA).rad
    c = cA.distanceTo(cB).rad
    A = cA.angleBetween(cB,cC).rad
    B = cB.angleBetween(cC,cA).rad
    C = cC.angleBetween(cA,cB).rad
    E = cA.area(cB,cC)

    uA, vA = center.project(cA, projection='lambert')
    uB, vB = center.project(cB, projection='lambert')
    uC, vC = center.project(cC, projection='lambert')

    # The shoelace formula gives the area of a triangle given coordinates:
    # A = 1/2 abs( (x2-x1)*(y3-y1) - (x3-x1)*(y2-y1) )
    area = 0.5 * abs( (uB.rad-uA.rad) * (vC.rad-vA.rad) - (uC.rad-uA.rad) * (vB.rad-vA.rad) )
    print('lambert area = ',area,E)
    np.testing.assert_allclose(area, E, rtol=1.e-8, err_msg="lambert didn't preserve area")

    # Check that project_rad does the same thing
    uA2, vA2 = center.project_rad(cA.ra.rad, cA.dec.rad, projection='lambert')
    np.testing.assert_allclose([uA2,vA2], [uA.rad,vA.rad], rtol=1.e-8,
                               err_msg="project_rad not equivalent")

    # Check the deprojection
    cA2 = center.deproject(uA, vA, projection='lambert')
    np.testing.assert_allclose(cA2.rad, cA.rad, err_msg="deproject didn't return to orig")
    cA3 = center.deproject_rad(uA.rad, vA.rad, projection='lambert')
    np.testing.assert_allclose(cA3, cA.rad, err_msg="deproject_rad not equivalent")
    check_aux(cA2)

    # The angles are not preserved
    a = sqrt( (uB.rad-uC.rad)**2 + (vB.rad-vC.rad)**2 )
    b = sqrt( (uC.rad-uA.rad)**2 + (vC.rad-vA.rad)**2 )
    c = sqrt( (uA.rad-uB.rad)**2 + (vA.rad-vB.rad)**2 )
    cosA = ((uB.rad-uA.rad)*(uC.rad-uA.rad) + (vB.rad-vA.rad)*(vC.rad-vA.rad)) / (b*c)
    cosB = ((uC.rad-uB.rad)*(uA.rad-uB.rad) + (vC.rad-vB.rad)*(vA.rad-vB.rad)) / (c*a)
    cosC = ((uA.rad-uC.rad)*(uB.rad-uC.rad) + (vA.rad-vC.rad)*(vB.rad-vC.rad)) / (a*b)

    print('lambert cosA = ',cosA,cos(A)) # 0.385002890444 0.371329625008
    print('lambert cosB = ',cosB,cos(B)) # 0.13803694316 0.102895820012
    print('lambert cosC = ',cosC,cos(C)) # 0.860935750473 0.885364487706
    assert cosA > cos(A) + 0.01
    assert cosB > cos(B) + 0.03
    assert cosC < cos(C) - 0.02

    # The deproject jacobian should tell us how the area changes
    dudx, dudy, dvdx, dvdy = center.jac_deproject(uA, vA, 'lambert').ravel()
    jac_area = abs(dudx*dvdy - dudy*dvdx)
    np.testing.assert_allclose(jac_area, E/area, err_msg='jac_deproject gave wrong area')

    dudx, dudy, dvdx, dvdy = center.jac_deproject_rad(uA.rad, vA.rad, 'lambert').ravel()
    np.testing.assert_allclose(jac_area, abs(dudx*dvdy - dudy*dvdx),
                               err_msg='jac_deproject_rad not equivalent')

    # center projects to 0,0 with unit area
    u, v = center.project(center, 'lambert')
    np.testing.assert_allclose([u.rad, v.rad], 0., err_msg='center did not project to (0,0)')
    c2 = center.deproject(u,v, 'lambert')
    np.testing.assert_allclose(c2.rad, center.rad, err_msg='(0,0) did not deproject to center')
    np.testing.assert_allclose(np.linalg.det(center.jac_deproject(u, v, 'lambert')), 1.,
                               err_msg='determinant of jac_deproject matrix != 1 at center')
    check_aux(c2)


@timer
def test_postel_projection():
    """Test the postel projection.
    """
    # Test that a small triangle has the correct properties.
    # Postel projections preserve distance from the center.
    center = coord.CelestialCoord(0.234 * radians, 0.342 * radians)
    cA = coord.CelestialCoord(-0.193 * radians, 0.882 * radians)
    cB = coord.CelestialCoord((-0.193 + 1.7e-8) * radians,
                               (0.882 + 1.2e-8) * radians)
    cC = coord.CelestialCoord((-0.193 - 2.4e-8) * radians,
                               (0.882 + 3.1e-8) * radians)

    a = cB.distanceTo(cC).rad
    b = cC.distanceTo(cA).rad
    c = cA.distanceTo(cB).rad
    A = cA.angleBetween(cB,cC).rad
    B = cB.angleBetween(cC,cA).rad
    C = cC.angleBetween(cA,cB).rad
    E = cA.area(cB,cC)

    uA, vA = center.project(cA, projection='postel')
    uB, vB = center.project(cB, projection='postel')
    uC, vC = center.project(cC, projection='postel')

    dA = sqrt( uA.rad**2 + vA.rad**2 )
    dB = sqrt( uB.rad**2 + vB.rad**2 )
    dC = sqrt( uC.rad**2 + vC.rad**2 )
    print('postel dA = ',dA,center.distanceTo(cA))
    print('postel dB = ',dB,center.distanceTo(cB))
    print('postel dC = ',dC,center.distanceTo(cC))
    np.testing.assert_allclose(dA, center.distanceTo(cA).rad)
    np.testing.assert_allclose(dB, center.distanceTo(cB).rad)
    np.testing.assert_allclose(dC, center.distanceTo(cC).rad)

    # Check that project_rad does the same thing
    uA2, vA2 = center.project_rad(cA.ra.rad, cA.dec.rad, projection='postel')
    np.testing.assert_allclose([uA2,vA2], [uA.rad,vA.rad], rtol=1.e-8,
                               err_msg="project_rad not equivalent")

    # Check the deprojection
    cA2 = center.deproject(uA, vA, projection='postel')
    np.testing.assert_allclose(cA2.rad, cA.rad, err_msg="deproject didn't return to orig")
    cA3 = center.deproject_rad(uA.rad, vA.rad, projection='postel')
    np.testing.assert_allclose(cA3, cA.rad, err_msg="deproject_rad not equivalent")
    check_aux(cA2)

    # The angles are not preserved
    a = sqrt( (uB.rad-uC.rad)**2 + (vB.rad-vC.rad)**2 )
    b = sqrt( (uC.rad-uA.rad)**2 + (vC.rad-vA.rad)**2 )
    c = sqrt( (uA.rad-uB.rad)**2 + (vA.rad-vB.rad)**2 )
    cosA = ((uB.rad-uA.rad)*(uC.rad-uA.rad) + (vB.rad-vA.rad)*(vC.rad-vA.rad)) / (b*c)
    cosB = ((uC.rad-uB.rad)*(uA.rad-uB.rad) + (vC.rad-vB.rad)*(vA.rad-vB.rad)) / (c*a)
    cosC = ((uA.rad-uC.rad)*(uB.rad-uC.rad) + (vA.rad-vC.rad)*(vB.rad-vC.rad)) / (a*b)

    print('postel cosA = ',cosA,cos(A))
    print('postel cosB = ',cosB,cos(B))
    print('postel cosC = ',cosC,cos(C))

    # The area is not preserved
    area = 0.5 * abs( (uB.rad-uA.rad) * (vC.rad-vA.rad) - (uC.rad-uA.rad) * (vB.rad-vA.rad) )
    print('postel area = ',area,E)

    # The deproject jacobian should tell us how the area changes
    dudx, dudy, dvdx, dvdy = center.jac_deproject(uA, vA, 'postel').ravel()
    jac_area = abs(dudx*dvdy - dudy*dvdx)
    np.testing.assert_allclose(jac_area, E/area, err_msg='jac_deproject gave wrong area')

    dudx, dudy, dvdx, dvdy = center.jac_deproject_rad(uA.rad, vA.rad, 'postel').ravel()
    np.testing.assert_allclose(jac_area, abs(dudx*dvdy - dudy*dvdx),
                               err_msg='jac_deproject_rad not equivalent')

    # center projects to 0,0 with unit area
    u, v = center.project(center, 'postel')
    np.testing.assert_allclose([u.rad, v.rad], 0., err_msg='center did not project to (0,0)')
    c2 = center.deproject(u,v, 'postel')
    np.testing.assert_allclose(c2.rad, center.rad, err_msg='(0,0) did not deproject to center')
    np.testing.assert_allclose(np.linalg.det(center.jac_deproject(u, v, 'postel')), 1.,
                               err_msg='determinant of jac_deproject matrix != 1 at center')
    check_aux(c2)


@timer
def test_precess():
    """Test precession between epochs.
    """
    # I don't have much of a test here.  The formulae are what they are.
    # But it should at least be the case that a precession trip that ends up
    # back at the original epoch should leave the coord unchanged.
    orig = coord.CelestialCoord(0.234 * radians, 0.342 * radians)

    c1 = orig.precess(2000., 1950.)
    c2 = c1.precess(1950., 1900.)
    c3 = c2.precess(1900., 2000.)
    np.testing.assert_almost_equal(c3.ra.rad, orig.ra.rad, decimal=10)
    np.testing.assert_almost_equal(c3.dec.rad, orig.dec.rad, decimal=10)
    check_aux(c1)
    check_aux(c2)
    check_aux(c3)

    # The no op is exact.
    c4 = orig.precess(2000., 2000.)
    assert c4 == orig
    check_aux(c4)

    # I found a website that does precession calculations, so check that we are
    # consistent with them.
    # http://www.bbastrodesigns.com/coordErrors.html
    dra_1950 = -(2. + 39.07/60.)/60. * hours / radians
    ddec_1950 = -(16. + 16.3/60.)/60. * degrees / radians
    print('delta from website: ',dra_1950,ddec_1950)
    print('delta from precess: ',(c1.ra-orig.ra),(c1.dec-orig.dec))
    np.testing.assert_almost_equal(dra_1950, c1.ra.rad-orig.ra.rad, decimal=6)
    np.testing.assert_almost_equal(ddec_1950, c1.dec.rad-orig.dec.rad, decimal=6)

    dra_1900 = -(5. + 17.74/60.)/60. * hours / radians
    ddec_1900 = -(32. + 35.4/60.)/60. * degrees / radians
    print('delta from website: ',dra_1900,ddec_1900)
    print('delta from precess: ',(c2.ra-orig.ra),(c2.dec-orig.dec))
    np.testing.assert_almost_equal(dra_1900, c2.ra.rad-orig.ra.rad, decimal=6)
    np.testing.assert_almost_equal(ddec_1900, c2.dec.rad-orig.dec.rad, decimal=6)


@timer
def test_galactic():
    """Test the conversion from equatorial to galactic coordinates and back.
    """
    # According to wikipedia: http://en.wikipedia.org/wiki/Galactic_coordinate_system
    # the galactic center is located at 17h:45.6m, -28.94d
    # But I get more precise values from https://arxiv.org/pdf/1010.3773.pdf
    center = coord.CelestialCoord(
        coord.Angle.from_hms('17:45:37.1991'),
        coord.Angle.from_dms('-28:56:10.2207'))
    print('center.galactic = ',center.galactic())
    el,b = center.galactic()
    np.testing.assert_almost_equal(el.wrap().rad, 0., decimal=8)
    np.testing.assert_almost_equal(b.rad, 0., decimal=8)

    # Go back from galactic coords to CelestialCoord
    center2 = coord.CelestialCoord.from_galactic(el,b)
    np.testing.assert_allclose(center2.rad, center.rad)
    check_aux(center2)

    # The north pole is at 12h:51.4m, 27.13d again with more precise values from the above paper.
    north = coord.CelestialCoord(
        coord.Angle.from_hms('12:51:26.27549'),
        coord.Angle.from_dms('27:07:41.7043'))
    print('north.galactic = ',north.galactic())
    el,b = north.galactic()
    np.testing.assert_allclose(b.rad, pi/2.)
    north2 = coord.CelestialCoord.from_galactic(el,b)
    np.testing.assert_allclose(north2.rad, north.rad)
    check_aux(north2)

    south = coord.CelestialCoord(
        coord.Angle.from_hms('00:51:26.27549'),
        coord.Angle.from_dms('-27:07:41.7043'))
    print('south.galactic = ',south.galactic())
    el,b = south.galactic()
    np.testing.assert_allclose(b.rad, -pi/2.)
    south2 = coord.CelestialCoord.from_galactic(el,b)
    np.testing.assert_allclose(south2.rad, south.rad)
    check_aux(south2)

    anticenter = coord.CelestialCoord(
        coord.Angle.from_hms('05:45:37.1991'),
        coord.Angle.from_dms('28:56:10.2207'))
    print('anticenter.galactic = ',anticenter.galactic())
    el,b = anticenter.galactic()
    np.testing.assert_almost_equal(el.rad, pi, decimal=8)
    np.testing.assert_almost_equal(b.rad, 0., decimal=8)
    anticenter2 = coord.CelestialCoord.from_galactic(el,b)
    np.testing.assert_allclose(anticenter2.rad, anticenter.rad)
    check_aux(anticenter2)


@timer
def test_ecliptic():
    """Test the conversion from equatorial to ecliptic coordinates and back.
    """
    # Use locations of ecliptic poles from http://en.wikipedia.org/wiki/Ecliptic_pole
    north_pole = coord.CelestialCoord(
        coord.Angle.from_hms('18:00:00.00'),
        coord.Angle.from_dms('66:33:38.55'))
    el, b = north_pole.ecliptic()
    # North pole should have b=90 degrees, with el being completely arbitrary.
    np.testing.assert_almost_equal(b.rad, pi/2, decimal=6)

    # Check round-trip
    north_pole2 = coord.CelestialCoord.from_ecliptic(el, b)
    np.testing.assert_almost_equal(north_pole2.ra.rad, north_pole.ra.rad, decimal=8)
    np.testing.assert_almost_equal(north_pole2.dec.rad, north_pole.dec.rad, decimal=8)
    check_aux(north_pole2)

    south_pole = coord.CelestialCoord(
        coord.Angle.from_hms('06:00:00.00'),
        coord.Angle.from_dms('-66:33:38.55'))
    el, b = south_pole.ecliptic()
    np.testing.assert_almost_equal(b.rad, -pi/2, decimal=6)
    south_pole2 = coord.CelestialCoord.from_ecliptic(el, b)
    np.testing.assert_almost_equal(south_pole2.ra.rad, south_pole.ra.rad, decimal=8)
    np.testing.assert_almost_equal(south_pole2.dec.rad, south_pole.dec.rad, decimal=8)
    check_aux(south_pole2)

    # Also confirm that positions that should be the same in equatorial and ecliptic coordinates are
    # actually the same:
    vernal = coord.CelestialCoord(0.*radians, 0.*radians)
    el, b = vernal.ecliptic()
    np.testing.assert_almost_equal(b.rad, 0., decimal=6)
    np.testing.assert_almost_equal(el.rad, 0., decimal=6)
    vernal2 = coord.CelestialCoord.from_ecliptic(el,b)
    np.testing.assert_almost_equal(vernal2.ra.rad, vernal.ra.rad, decimal=8)
    np.testing.assert_almost_equal(vernal2.dec.rad, vernal.dec.rad, decimal=8)
    check_aux(vernal2)

    autumnal = coord.CelestialCoord(pi*radians, 0.*radians)
    el, b = autumnal.ecliptic()
    np.testing.assert_almost_equal(el.wrap(pi*radians).rad, pi, decimal=6)
    np.testing.assert_almost_equal(b.rad, 0., decimal=6)
    autumnal2 = coord.CelestialCoord.from_ecliptic(el,b)
    np.testing.assert_almost_equal(autumnal2.ra.rad, autumnal.ra.rad, decimal=8)
    np.testing.assert_almost_equal(autumnal2.dec.rad, autumnal.dec.rad, decimal=8)
    check_aux(autumnal2)


@timer
def test_ecliptic_date():
    """Test the date option of the ecliptic and from_ecliptic functions.
    """
    # Test the results of using a date to get ecliptic coordinates with respect to the sun,
    # instead of absolute ones. For this, use dates and times of vernal and autumnal equinox
    # in 2014 from
    #   http://wwp.greenwichmeantime.com/longest-day/
    # and the conversion to Julian dates from
    #   http://www.aavso.org/jd-calculator

    import datetime
    vernal_eq_date = datetime.datetime(2014,3,20,16,57,0)
    vernal = coord.CelestialCoord(0.*radians, 0.*radians)
    el, b = vernal.ecliptic(epoch=2014)
    el_rel, b_rel = vernal.ecliptic(epoch=2014, date=vernal_eq_date)
    # Vernal equinox: should have (el, b) = (el_rel, b_rel) = 0.0
    np.testing.assert_almost_equal(el.rad, el_rel.rad, decimal=4)
    np.testing.assert_almost_equal(b.rad, b_rel.rad, decimal=6)
    vernal2 = coord.CelestialCoord.from_ecliptic(el_rel, b_rel, date=vernal_eq_date)
    np.testing.assert_almost_equal(vernal2.ra.wrap().rad, vernal.ra.rad, decimal=8)
    np.testing.assert_almost_equal(vernal2.dec.rad, vernal.dec.rad, decimal=8)
    check_aux(vernal2)

    # Now do the autumnal equinox: should have (el, b) = (pi, 0) = (el_rel, b_rel) when we look at
    # the time of the vernal equinox.
    autumnal = coord.CelestialCoord(pi*radians, 0.*radians)
    el, b = autumnal.ecliptic(epoch=2014)
    el_rel, b_rel = autumnal.ecliptic(epoch=2014, date=vernal_eq_date)
    np.testing.assert_almost_equal(el_rel.wrap(pi*radians).rad, el.wrap(pi*radians).rad,
                                   decimal=4)
    np.testing.assert_almost_equal(b_rel.rad, b.rad, decimal=6)
    autumnal2 = coord.CelestialCoord.from_ecliptic(el_rel, b_rel, date=vernal_eq_date)
    np.testing.assert_almost_equal(autumnal2.ra.wrap(pi*radians).rad,
                                   autumnal.ra.wrap(pi*radians).rad, decimal=8)
    np.testing.assert_almost_equal(autumnal2.dec.rad, autumnal.dec.rad, decimal=8)
    check_aux(autumnal2)

    # And check that if it's the date of the autumnal equinox (sun at (180, 0)) but we're looking at
    # the position of the vernal equinox (0, 0), then (el_rel, b_rel) = (-180, 0)
    autumnal_eq_date = datetime.datetime(2014,9,23,2,29,0)
    el_rel, b_rel = vernal.ecliptic(epoch=2014, date=autumnal_eq_date)
    np.testing.assert_almost_equal(el_rel.wrap(-pi*radians).rad, -pi, decimal=4)
    np.testing.assert_almost_equal(b_rel.rad, 0., decimal=6)

    # And check that if it's the date of the vernal equinox (sun at (0, 0)) but we're looking at
    # the position of the autumnal equinox (180, 0), then (el_rel, b_rel) = (180, 0)
    el_rel, b_rel = autumnal.ecliptic(epoch=2014, date=vernal_eq_date)
    np.testing.assert_almost_equal(el_rel.wrap(pi*radians).rad, pi, decimal=4)
    np.testing.assert_almost_equal(b_rel.rad, 0., decimal=6)

if __name__ == '__main__':
    test_init()
    test_eq()
    test_distance()
    test_xyz()
    test_xyz_array()
    test_angleBetween()
    test_gnomonic_projection()
    test_stereographic_projection()
    test_lambert_projection()
    test_postel_projection()
    test_precess()
    test_galactic()
    test_ecliptic()
    test_ecliptic_date()



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import numpy as np
import math
import coord

from helper_util import *

# We'll use this a lot, so just import it.
from math import pi

@timer
def test_init():
    """Test the AngleUnit initializer
    """
    # This is from the AngleUnit doc string.  Make sure it works!
    gradians = coord.AngleUnit(2.*pi / 400.)
    assert gradians.value == 2.*pi / 400., 'gradians value not 2pi/400'

    # Can also use named keyword argument
    np.testing.assert_almost_equal(coord.AngleUnit(value=17).value, 17, decimal=12)

    # Other types are ok as the value argument so long as they are convertible to float.
    np.testing.assert_almost_equal(coord.AngleUnit(np.float64(0.17)).value, 0.17, decimal=12,
                                   err_msg='using np.float64 for value failed')
    np.testing.assert_almost_equal(coord.AngleUnit(np.float32(0.23)).value, 0.23, decimal=8,
                                   err_msg='using np.float32 for value failed')
    np.testing.assert_almost_equal(coord.AngleUnit('1.7').value, 1.7, decimal=12,
                                   err_msg='using string "1.7" for value failed')


@timer
def test_builtin_units():
    """Check the built-in AngleUnits
    """
    np.testing.assert_almost_equal(coord.radians.value, 1., decimal=12)
    np.testing.assert_almost_equal(coord.degrees.value, pi / 180., decimal=12)
    np.testing.assert_almost_equal(coord.hours.value, pi / 12., decimal=12)
    np.testing.assert_almost_equal(coord.arcmin.value, pi / 180. / 60., decimal=12)
    np.testing.assert_almost_equal(coord.arcsec.value, pi / 180. / 3600., decimal=12)


@timer
def test_invalid():
    """Check that invalid constructors raise the appropriate exceptions.
    """
    # Wrong type of value argument
    np.testing.assert_raises(TypeError, coord.AngleUnit, coord.degrees)
    # Also wrong type, but strings give a ValueError
    np.testing.assert_raises(ValueError, coord.AngleUnit, 'spam')
    # Wrong number of arguments
    np.testing.assert_raises(TypeError, coord.AngleUnit, 1, 3)
    np.testing.assert_raises(TypeError, coord.AngleUnit)
    # Wrong keyword argument
    np.testing.assert_raises(TypeError, coord.AngleUnit, the_value=0.2)


@timer
def test_div():
    """Test AngleUnit / AngleUnit
    """
    np.testing.assert_almost_equal(coord.degrees / coord.arcmin, 60., decimal=12)
    np.testing.assert_almost_equal(coord.degrees / coord.arcsec, 3600., decimal=12)
    np.testing.assert_almost_equal(coord.hours / coord.degrees, 15., decimal=12)
    np.testing.assert_almost_equal(coord.hours / coord.hours, 1., decimal=12)

    # Dividing by radians is just the value attribute.
    np.testing.assert_almost_equal(coord.degrees / coord.radians, coord.degrees.value, decimal=12)
    np.testing.assert_almost_equal(coord.arcsec / coord.radians, coord.arcsec.value, decimal=12)

    # We claim this is equivalent to (1 * AngleUnit) / AngleUnit.  Check.
    np.testing.assert_almost_equal((1. * coord.degrees) / coord.arcmin, 60., decimal=12)
    np.testing.assert_almost_equal((1. * coord.degrees) / coord.arcsec, 3600., decimal=12)

    # In general, it is associative.
    np.testing.assert_almost_equal((23.17 * coord.hours) / coord.radians,
                                   23.17 * (coord.hours / coord.radians), decimal=12)

    # Invalid to divide by something other than an AngleUnit
    np.testing.assert_raises(TypeError, coord.degrees.__div__, 12)
    np.testing.assert_raises(TypeError, coord.degrees.__div__, 12 * coord.arcsec)
    np.testing.assert_raises(TypeError, coord.degrees.__div__, 'arcmin')


@timer
def test_pickle():
    """Check that AngleUnits pickle correctly.

    Not a big deal here, but if we start including any C structs, they require special handling
    to make picklable, so useful to have this check in place.

    More usefully, the do_pickle function also checks that eval(repr(obj)) returns something
    equivalent to the original object along with a few other sanity checks such as obj != None.
    So it's a good idea to just run this on every kind of object we have.
    """
    do_pickle(coord.radians)
    do_pickle(coord.degrees)
    do_pickle(coord.hours)
    do_pickle(coord.arcmin)
    do_pickle(coord.arcsec)
    gradians = coord.AngleUnit(2.*pi / 400.)
    do_pickle(gradians)


@timer
def test_eq():
    """Check that equal units are equal, but unequal ones are not.
    """
    # The alt versions should be equal to the built-in versions.
    alt_radians = coord.AngleUnit(1.)
    alt_degrees = coord.AngleUnit(pi/180.)
    assert coord.radians == alt_radians
    assert coord.degrees == alt_degrees

    # These should all test as unequal.  Note some non-AngleUnits in the list.
    diff_list = [ coord.radians, alt_degrees, coord.hours, coord.arcmin, coord.arcsec,
                  1.0, coord.AngleUnit, None ]
    all_obj_diff(diff_list)


@timer
def test_from_name():
    """Test the AngleUnit from_name static method

    Note in particular that shorter or longer versions are allowed.  As are caps or lowercase.
    """
    assert coord.AngleUnit.from_name('radians') == coord.radians
    assert coord.AngleUnit.from_name('Radian') == coord.radians
    assert coord.AngleUnit.from_name('rad') == coord.radians
    assert coord.AngleUnit.from_name('RADians FTW') == coord.radians  # unlikely usage, but valid.
    assert coord.AngleUnit.from_name('degrees') == coord.degrees
    assert coord.AngleUnit.from_name('degree') == coord.degrees
    assert coord.AngleUnit.from_name('degr') == coord.degrees
    assert coord.AngleUnit.from_name('DEG') == coord.degrees
    assert coord.AngleUnit.from_name('hours') == coord.hours
    assert coord.AngleUnit.from_name('hour') == coord.hours
    assert coord.AngleUnit.from_name('hrs') == coord.hours
    assert coord.AngleUnit.from_name('hr') == coord.hours
    assert coord.AngleUnit.from_name('arcmin') == coord.arcmin
    assert coord.AngleUnit.from_name('arcminutes') == coord.arcmin
    assert coord.AngleUnit.from_name('arcsec') == coord.arcsec
    assert coord.AngleUnit.from_name('arcseconds') == coord.arcsec

    np.testing.assert_raises(ValueError, coord.AngleUnit.from_name, 'gradians')
    np.testing.assert_raises(ValueError, coord.AngleUnit.from_name, 'spam')

    # Make sure all the angles in AngleUnit.valid_names work
    for name in coord.AngleUnit.valid_names:
        assert isinstance(coord.AngleUnit.from_name(name), coord.AngleUnit)

if __name__ == '__main__':
    test_init()
    test_builtin_units()
    test_invalid()
    test_div()
    test_pickle()
    test_eq()
    test_from_name()



# Copyright (c) 2013-2017 LSST Dark Energy Science Collaboration (DESC)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This file tests the equivalence of the CelestialCoord functionality with the corresponding
# functionality in the astropy package.

from __future__ import print_function
import numpy as np
import math
import time
import coord
import astropy.coordinates
from astropy import units

from helper_util import *

# We'll use these a lot, so just import them.
from math import pi

@timer
def test_angle():
    """Tests comparing coord.Angle to astropy.coordinates.Angle
    """
    # radians
    theta_coord = 45. * coord.degrees
    theta_astro = astropy.coordinates.Angle(pi/4., units.radian)

    # degrees
    np.testing.assert_almost_equal(theta_coord.rad, theta_astro.rad, decimal=12)
    np.testing.assert_almost_equal(theta_coord / coord.degrees, theta_astro.degree, decimal=12)
    np.testing.assert_almost_equal(theta_coord / coord.hours, theta_astro.hour, decimal=12)
    np.testing.assert_almost_equal(theta_coord / coord.arcmin, theta_astro.arcminute, decimal=12)
    np.testing.assert_almost_equal(theta_coord / coord.arcsec, theta_astro.arcsec, decimal=12)

    # Other constructors
    theta_astro2 = astropy.coordinates.Angle(23.09, units.arcsec)
    theta_coord2 = coord.Angle(23.09, coord.arcsec)
    np.testing.assert_almost_equal(theta_coord2.rad, theta_astro2.rad, decimal=12)

    theta_astro3 = astropy.coordinates.Angle(-0.17, unit='rad')
    theta_coord3 = coord._Angle(-0.17)
    np.testing.assert_almost_equal(theta_coord3.rad, theta_astro3.rad, decimal=12)

    # astropy wrapping uses a different convention than we do.  Their argument is
    # the upper end of the target range, not the center.
    theta_astro4 = theta_astro3.wrap_at(360 * units.deg)
    theta_coord4 = theta_coord3.wrap(180 * coord.degrees)
    np.testing.assert_almost_equal(theta_coord4.rad, theta_astro4.rad, decimal=12)

    theta_astro5 = theta_astro3.wrap_at(-100 * units.deg)
    theta_coord5 = theta_coord3.wrap(-280 * coord.degrees)
    np.testing.assert_almost_equal(theta_coord5.rad, theta_astro5.rad, decimal=12)

    theta_astro6 = astropy.coordinates.Angle('03:34:12', unit='hourangle')
    theta_coord6 = coord.Angle.from_hms('03:34:12')
    np.testing.assert_almost_equal(theta_coord6.rad, theta_astro6.rad, decimal=12)

    theta_astro7 = astropy.coordinates.Angle('03:34:12', unit='deg')
    theta_coord7 = coord.Angle.from_dms('03:34:12')
    np.testing.assert_almost_equal(theta_coord7.rad, theta_astro7.rad, decimal=12)

    # Their default arguments to to_string are different from ours, but can make them compatible.
    print('theta_astro6.hms = ',theta_astro6.to_string(sep=':', pad=True))
    print('theta_coord6.hms = ',theta_coord6.hms())
    assert theta_coord6.hms() == theta_astro6.to_string(sep=':', pad=True)

    print('theta_astro7.dms = ',theta_astro7.to_string(sep=':', pad=True))
    print('theta_coord7.dms = ',theta_coord7.dms())
    assert theta_coord7.dms() == theta_astro7.to_string(sep=':', pad=True)

    print('theta_astro6.hms = ',theta_astro6.to_string())
    print('theta_coord6.hms = ',theta_coord6.hms(sep='hms', pad=False))
    assert theta_coord6.hms(sep='hms', pad=False) == theta_astro6.to_string()

    print('theta_astro7.hms = ',theta_astro7.to_string())
    print('theta_coord7.hms = ',theta_coord7.dms(sep='dms', pad=False))
    assert theta_coord7.dms(sep='dms', pad=False) == theta_astro7.to_string()


@timer
def test_basic():
    """Basic tests of CelestialCoord construction and corresponding SkyCoord construction.
    """
    a1 = astropy.coordinates.SkyCoord(10.625 * units.degree, 41.2 * units.degree)
    c1 = coord.CelestialCoord(10.625 * coord.degrees, 41.2 * coord.degrees)
    np.testing.assert_almost_equal(c1.rad, [a1.ra.rad, a1.dec.rad], decimal=12)

    a2 = astropy.coordinates.SkyCoord('00:42.5 +41:12', unit=(units.hourangle, units.deg))
    c2 = coord.CelestialCoord(coord.Angle.from_hms('00:42:30'), coord.Angle.from_dms('41:12:00'))
    np.testing.assert_almost_equal(c2.rad, [a2.ra.rad, a2.dec.rad], decimal=12)

    # astropy made different choices for the HMS and DMS string representations, but we can
    # make it match our choice without too much trouble.
    astr = a2.to_string(style='hmsdms', sep=':')
    print('astr = ',astr)
    astr_ra, astr_dec = astr.split()
    cstr_ra = c2.ra.hms()
    cstr_dec = c2.dec.dms(plus_sign=True)
    assert astr_ra == cstr_ra
    assert astr_dec == cstr_dec


@timer
def test_distance():
    """Test calculations of distances on the sphere.
    """
    t0 = time.time()
    c1 = coord.CelestialCoord(0.234 * coord.radians, 0.342 * coord.radians)
    c2 = coord.CelestialCoord(0.234 * coord.radians, -1.093 * coord.radians)
    c3 = coord.CelestialCoord((pi + 0.234) * coord.radians, -0.342 * coord.radians)
    c4 = coord.CelestialCoord((pi + 0.234) * coord.radians, 0.832 * coord.radians)
    c5 = coord.CelestialCoord(1.832 * coord.radians, -0.723 * coord.radians)
    c6 = coord.CelestialCoord((0.234 + 2.3e-9) * coord.radians, (0.342 + 1.2e-9) * coord.radians)
    t1 = time.time()

    a1 = astropy.coordinates.SkyCoord(0.234 * units.radian, 0.342 * units.radian)
    a2 = astropy.coordinates.SkyCoord(0.234 * units.radian, -1.093 * units.radian)
    a3 = astropy.coordinates.SkyCoord((pi + 0.234) * units.radian, -0.342 * units.radian)
    a4 = astropy.coordinates.SkyCoord((pi + 0.234) * units.radian, 0.832 * units.radian)
    a5 = astropy.coordinates.SkyCoord(1.832 * units.radian, -0.723 * units.radian)
    a6 = astropy.coordinates.SkyCoord(0.234 + 2.3e-9, 0.342 + 1.2e-9, unit=units.radian)
    t2 = time.time()

    coord_dist = [c1.distanceTo(c).rad for c in [c2,c3,c4,c5,c6]]
    t3 = time.time()
    astropy_dist = [a1.separation(a).rad for a in [a2,a3,a4,a5,a6]]
    t4 = time.time()

    np.testing.assert_almost_equal(coord_dist, astropy_dist, decimal=12)
    # For the last one, the distance is rather small in radians, so test in arcsec
    np.testing.assert_almost_equal(coord_dist[-1] * (coord.radians/coord.arcsec),
                                   astropy_dist[-1] * (coord.radians/coord.arcsec), decimal=10)

    print('Compare times for distance calculations:')
    print('  Make CelestialCoords: t = ',t1-t0)
    print('  Make SkyCoords: t = ',t2-t1)
    print('  Calculate distances with Coord: t = ',t3-t2)
    print('  Calculate distances with Astropy: t = ',t4-t3)


@timer
def test_precess():
    """Test precession between epochs.
    """
    t0 = time.time()
    c1 = coord.CelestialCoord(0.234 * coord.radians, 0.342 * coord.radians)
    t1 = time.time()
    c2 = c1.precess(2000, 1950)
    c3 = c2.precess(1950, 1900)
    t2 = time.time()

    a1 = astropy.coordinates.SkyCoord(0.234, 0.342, unit=units.radian,
                                      frame=astropy.coordinates.FK5(equinox='J2000'))
    t3 = time.time()
    a2 = a1.transform_to(astropy.coordinates.FK5(equinox='J1950'))
    a3 = a2.transform_to(astropy.coordinates.FK5(equinox='J1900'))
    t4 = time.time()

    np.testing.assert_allclose(c1.rad, [a1.ra.rad, a1.dec.rad], rtol=1.e-5,
                               err_msg='starting coords different')
    np.testing.assert_allclose(c2.rad, [a2.ra.rad, a2.dec.rad], rtol=1.e-5,
                               err_msg='coord/astropy different after 2000->1950')
    np.testing.assert_allclose(c3.rad, [a3.ra.rad, a3.dec.rad], rtol=1.e-5,
                               err_msg='coord/astropy different after 1950->1900')

    print('Compare times for precession calculations:')
    print('  Make CelestialCoords: t = ',t1-t0)
    print('  Make SkyCoords: t = ',t3-t2)
    print('  Precess with Coord: t = ',t2-t1)
    print('  Precess with Astropy: t = ',t4-t3)
    # On my laptop, these times are
    #   Make CelestialCoords: t =  9.10758972168e-05
    #   Make SkyCoords: t =  0.00361394882202
    #   Precess with Coord: t =  0.000560998916626
    #   Precess with Astropy: t =  0.0377740859985

    # Make sure we don't get slow like astropy.  ;)
    # (Travis is a bit slower than the above times, but these limits should still be safe.)
    assert t1-t0 < 0.001, 'Building CelestialCoord is too slow'
    assert t2-t1 < 0.01, 'CelestialCoord.precess is too slow'


@timer
def test_galactic():
    """Test the conversion from equatorial to galactic coordinates and back.
    """
    center = coord.CelestialCoord(coord.Angle.from_hms('17:45:37.1991'),
                                  coord.Angle.from_dms('-28:56:10.2207'))
    north = coord.CelestialCoord(coord.Angle.from_hms('12:51:26.27549'),
                                 coord.Angle.from_dms('27:07:41.7043'))
    south = coord.CelestialCoord(coord.Angle.from_hms('00:51:26.27549'),
                                 coord.Angle.from_dms('-27:07:41.7043'))
    anticenter = coord.CelestialCoord(coord.Angle.from_hms('05:45:37.1991'),
                                      coord.Angle.from_dms('28:56:10.2207'))
    random = coord.CelestialCoord(0.234 * coord.radians, 0.342 * coord.radians)

    for c1 in [center, north, south, anticenter, random]:
        a1 = astropy.coordinates.SkyCoord(*c1.rad, unit=units.rad, frame='fk5')
        #print('c1.galactic() = ',c1.galactic())
        #print('a1.galactic = ',a1.galactic)
        el, b = c1.galactic()
        # Wrap el to the same phase as a1
        el = el.wrap(a1.galactic.l.rad * coord.radians)
        if c1 not in [north, south]:
            np.testing.assert_almost_equal(el.rad, a1.galactic.l.rad, decimal=6)
        np.testing.assert_almost_equal(b.rad, a1.galactic.b.rad, decimal=6)

        c2 = coord.CelestialCoord.from_galactic(el,b)
        a2 = astropy.coordinates.SkyCoord(el.rad, b.rad, unit=units.radian, frame='galactic')
        a2 = a2.transform_to('fk5')
        c2_ra = c2.ra.wrap(a2.ra.rad * coord.radians)
        np.testing.assert_almost_equal(c2_ra.rad, a2.ra.rad, decimal=6)
        np.testing.assert_almost_equal(c2.dec.rad, a2.dec.rad, decimal=6)

@timer
def test_ecliptic():
    """Test the conversion from equatorial to ecliptic coordinates and back.
    """
    # Note: These don't have the same level of agreement as galactic coords did.
    # Possibly this is related to a caveat in the astropy documentation:
    #     "In the current version of astropy, the ecliptic frames do not yet have stringent
    #      accuracy tests."
    # Or possibly, the Coord conversions aren't super accurate.  I'm not really sure.
    north = coord.CelestialCoord(coord.Angle.from_hms('18:00:00.00'),
                                 coord.Angle.from_dms('66:33:38.55'))
    south = coord.CelestialCoord(coord.Angle.from_hms('06:00:00.00'),
                                 coord.Angle.from_dms('-66:33:38.55'))
    vernal = coord.CelestialCoord(0.*coord.radians, 0.*coord.radians)
    autumnal = coord.CelestialCoord(pi*coord.radians, 0.*coord.radians)
    random = coord.CelestialCoord(0.234 * coord.radians, 0.342 * coord.radians)

    for c1 in [north, south, vernal, autumnal, random]:
        a1 = astropy.coordinates.SkyCoord(*c1.rad, unit=units.rad, frame='fk5')
        eclip =  a1.transform_to('geocentrictrueecliptic')
        #print('c1.ecliptic() = ',c1.ecliptic())
        #print('a1.ecliptic = ',eclip)
        el, b = c1.ecliptic()
        el = el.wrap(eclip.lon.rad * coord.radians)
        if c1 not in [north, south]:
            np.testing.assert_almost_equal(el.rad, eclip.lon.rad, decimal=4)
        np.testing.assert_almost_equal(b.rad, eclip.lat.rad, decimal=4)

        c2 = coord.CelestialCoord.from_ecliptic(el,b)
        a2 = astropy.coordinates.SkyCoord(el.rad, b.rad, unit=units.radian,
                                          frame='geocentrictrueecliptic')
        a2 = a2.transform_to('fk5')
        c2_ra = c2.ra.wrap(a2.ra.rad * coord.radians)
        np.testing.assert_almost_equal(c2_ra.rad, a2.ra.rad, decimal=3)
        np.testing.assert_almost_equal(c2.dec.rad, a2.dec.rad, decimal=3)

if __name__ == '__main__':
    test_angle()
    test_basic()
    test_distance()
    test_precess()
    test_galactic()
    test_ecliptic()
